# -*- coding: utf-8 -*-

import numpy as np

from ..uncertainty_basis import UncertaintyBasis

class Demands(UncertaintyBasis):
    
    def __init__(self, n_scenarios, n_commodities, distribution, **kwargs):
        self.n_scenarios = n_scenarios
        self.n_commodities = n_commodities
        self.distribution = distribution
        
        self._scenarios = None
        
        if self.distribution == "lognormal":
            self.mean = kwargs.get("mean", 0)
            self.var = kwargs.get("var", 1)
            self.corr = kwargs.get("corr", 0)
            self.label = "logN({},{})-corr{}".format(self.mean, self.var, self.corr)
            
        elif self.distribution == "uniform":
            self.lb = kwargs.get("lb", 0)
            self.ub = kwargs.get("ub", 1)
            self.corr = kwargs.get("corr", 0)
            self.label = "U[{},{}]-corr{}".format(self.lb, self.ub, self.corr)
            if self.lb >= self.ub:
                raise ValueError("Wrong lower and upper bound")
          
        self.randomize()
        
        UncertaintyBasis.__init__(self, 
                                  name=f"Commodity demand uncertainty: {self.label}", 
                                  n_scenarios=self.n_scenarios,
                                  features={'d': self.n_commodities})
        
    def randomize(self):
        if self.distribution == "lognormal":
            self._scenarios = np.array([self.lognormal() for i in range(self.n_scenarios)])
        elif self.distribution == "uniform":
            self._scenarios = np.array([self.uniform() for i in range(self.n_scenarios)])
    
    def get_scenario(self, scen_index):
        assert isinstance(scen_index, int), f"`scen_index should be of type int, not {type(scen_index)}."
        return {'d': self._scenarios[scen_index]}
    
    def lognormal(self):
        mu = [self.mean] * self.n_commodities
        sigma = [[self.corr if i!=j else self.var 
                      for i in range(self.n_commodities)] 
                        for j in range(self.n_commodities)]
        return np.exp(np.random.multivariate_normal(mu, sigma)).astype('int64')
    
    def uniform(self):
        return np.random.randint(self.lb, self.ub+1, size=self.n_commodities)



# -*- coding: utf-8 -*-

from typing import Callable, Optional

import numpy as np

from stochoptim.scengen.variability_process import VariabilityProcess
from stochoptim.scengen.tree_structure import Node


class FigureOfDemerit:
    
    def __init__(self, 
                 demerit_fct: Callable[[int, np.ndarray, np.ndarray], float],
                 variability_process: VariabilityProcess):
        """
        Arguments:
        ----------
        demerit_fct: Callable[[int, 2d-array, 1d-array], float]
            The function giving the quality value (so-called 'demerit') of the discretization points and weights.
            See the signature of the `sample_demerit` method.
            
        variability_process: VariabilityProcess
            The variability process used to compute the figure of demerit of a subtree.
        """
        self._demerit_fct = demerit_fct
        self._variability_process = variability_process
        
    def sample_demerit(self, 
                       stage: int, 
                       epsilons: np.ndarray, 
                       weights: np.ndarray) -> float:
        r"""Demerit (quality measure) of the discretization points and weights 
        of the stagewise independent random vector \epsilon_{stage}.
        
        Arguments:
        ----------
        stage: int >= 0
        
        epsilons: 2d-array
            The discretization points given as an array of shape (n_sample, dim_epsilon).
            
        weights: 1d-array
            The discretization weights given as an array of shape (n_sample,).
            
        disc_method: string (optional)
            The method used to generate the discretization points and weights
            
        Returns:
        --------
        float > 0: 
            Positive number representing the quality measure (so-called 'demerit') of the discretization 
            points and weights.
        """
        return self._demerit_fct(stage, epsilons, weights)

    def children_demerit(self, node: Node) -> float:
        """Demerit (quality measure) of the discretization points and weights at the child nodes of 'node'.
        
        Argument:
        ---------
        node: Node
        
        Returns:
        --------
        float > 0: value of the demerit.
        """
        if node.is_leaf:
            return 0
        else:
            return self.sample_demerit(stage=node.level, 
                                       epsilons=np.array([c.data.get("eps") for c in node.children]),
                                       weights=np.array([c.data["w"] for c in node.children]))
        
    def lookback_demerit(self, node: Node) -> float:
        """Figure of demerit as defined in [1, eq.(45)] and computed using the recursion [2, eq.(26)-(27)].
        
        This is the general formula for the figure of demerit, which can be computed only when the scenario tree 
        has been filled (optimally or not) with the scenarios at each node. It uses the path-dependent (lookback)
        variability function. 
        
        [1]: Keutchayan, Munger, Gendreau, Bastin (2018) 'A quality measure for the discretization of probability 
        distributions in multistage stochastic optimization'
        [2]: Keutchayan, Munger, Gendreau (2019) 'On the scenario-tree optimal-value error 
        for stochastic programming problems'
        
        Argument:
        ---------
        node: Node
            If node is the tree root, then it outputs the actual figure of demerit of the scenario tree.
            If node is not the root, then it computes only the subset of the whole sum over the nodes in the subtree
            rooted at 'node'.
        
        Returns:
        --------
        float > 0: value of the demerit.
        """
        if node.is_leaf:
            return 0
        else:
            return self.children_demerit(node) * self._variability_process.node_lookback(node) \
                        + sum(child.data["w"] * self.lookback_demerit(child) for child in node.children)
    
    def looknow_demerit(self, node: Node) -> float:
        """Figure of demerit as defined in [1, eq.(49)] and computed using the recursion [1, eq.(50)-(51)]. 
        
        This is the simplified formula that only features the path-independent (looknow) variability. It coincides 
        with the general formula (given by lookback demerit) if the variability is of the product form accross stages.
        
        [1]: Keutchayan, Munger, Gendreau, Bastin (2018) 'A quality measure for the discretization of probability 
        distributions in multistage stochastic optimization'.
        
        Argument:
        ---------
        node: Node
            If node is the tree root *and* variability is of the product form, then it outputs the actual 
            figure of demerit of the scenario tree (it then coincides with .lookback_demerit(node)).
            If node is not the root *and* variability is of the product form, then it computes figure of demerit 
            of the sub-scenario-tree rooted at 'node'.
        
        Returns:
        --------
        float > 0: value of the demerit.
        """
        if node.is_leaf:
            return 0
        else:
            return self.children_demerit(node) + sum(child.data["w"] 
                                                    * self._variability_process.node_looknow(child)
                                                    * self.looknow_demerit(child) for child in node.children)
    
    def __call__(self, 
                 node: Node, 
                 subtree: bool, 
                 path: Optional[bool] = None) -> float:
        """Returns the demerit of the node it is called on.
        
        Arguments:
        ----------
        node: Node
            The node for which the demerit will be computed.
            
        subtree: bool
            If True, the demerit of the whole subtree rooted at 'node' is computed, 
            otherwise only the demerit of its child nodes.
            
        path: bool or None
            If True, the path-dependent variability is used to compute the subtree demerit, 
            otherwise the path-independent is used. (Makes sense only if the *subtree* demerit is computed, 
            i.e., if subtree is True.)
            
        Returns:
        --------
        float: the quality measure (so-called 'demerit') of a node
        """
        if subtree:
            if path:
                return self.lookback_demerit(node)
            else:
                return self.looknow_demerit(node)
        else:
            return self.children_demerit(node)
            


# -*- coding: utf-8 -*-

from typing import Dict, Optional, Callable
import numpy as np

from stochoptim.scengen.tree_structure import get_data_path
from stochoptim.scengen.tree_structure import Node

ScenarioType = Dict[str, np.ndarray]
ScenarioPathType = Dict[int, ScenarioType]

class VariabilityProcess:
    """The process that maps scenario paths to variability values"""

    def __init__(self, 
                 lookback_fct: Optional[Callable[[int, ScenarioPathType], float]] = None, 
                 looknow_fct: Optional[Callable[[int, np.ndarray], float]] = None, 
                 average_fct: Optional[Callable[[int], float]] = None, 
                 name: Optional[str] = None,
                 checker: bool = True):
        """
        Arguments:
        ----------
        lookback_fct: Callable[[int, Dict[int, Dict[str, 1d-array]]], float] or None (default: None)
            See the signature of the `lookback` method.
            
        looknow_fct: Callable[[int, 1d-array], float] or None (default: None)
            See the signature of the `looknow` method.
            
        average_fct: Callable[[int], float] or None (default: None)
            See the signature of the `average` method.
            
        name: str or None (default: None)
            Name of the variability process
        """
        self._lookback_fct = lookback_fct
        self._looknow_fct = looknow_fct
        self._average_fct = average_fct
        self._name = name
        self._checker = checker
        
    def has_lookback(self) -> bool:
        return self._lookback_fct is not None

    def has_looknow(self) -> bool:
        return self._looknow_fct is not None
    
    def has_average(self) -> bool:
        return self._average_fct is not None
    
    def lookback(self, stage: int, scenario_path: ScenarioPathType) -> float:
        r""" Path-dependent variability function. 
        
        Mathematically it is the implementation of
            - V_0   (variability at stage 0)
            - V_{stage}(\xi_1, ..., \xi_{stage})    (variability at stage >= 1).
            
        Arguments:
        ----------           
        stage: int >= 0
            Current stage of the variability function to be computed (starting at stage 0).
            
        scenario_path: Dict[int, Dict[str, 1d-array]]
            History of the process up to and including the current stage.
                
        Returns:
        --------
        float >= 0: Value of variability.
        """
        varability = self._lookback_fct(stage, scenario_path)
        if self._checker:
            assert isinstance(varability, (int, float)) or varability.shape == (), \
            f"The `lookback_fct` function must return an int or float, not {type(varability)}"
        return varability
    
    def looknow(self, stage: int, epsilon: np.ndarray) -> float:
        r""" Path-independent variability function. 
        
        Mathematically it is the implementation of
            - V_0   (variability at stage 0)
            - V_{stage}(\epsilon_{stage})    (variability at stage >= 1).
            
        Arguments:
        ----------           
        stage: int >= 0
            Current stage of the variability function to be computed (starting at stage 0).
            
        epsilon: 1d-array
            Discretization point of \epsilon_{stage}, of shape (dim_epsilon,).
                
        Returns:
        --------
        float >= 0: Value of variability.
        """
        varability = self._looknow_fct(stage, epsilon)
        if self._checker:
            assert isinstance(varability, (int, float)) or varability.shape == (), \
            f"The `looknow_fct` function must return an int or float, not {type(varability)}"
        return varability
            
    def average(self, stage: int) -> float:
        r""" Average variability, defined as the expectation of the path-dependent variability under the scenario
        process. 
        
        Mathematically: 
            V_0 (variability at stage 0)
            E[V_{stage}(\xi_1, ..., \xi_{stage})] for stage = 1, ..., last_stage-1
            
        Argument:
        ---------
        stage: int >= 0
        
        Returns:
        --------
        float >= 0: Average variability
        """
        varability = self._average_fct(stage)
        if self._checker:
            assert isinstance(varability, (int, float)) or varability.shape == (), \
            f"The `average_fct` function must return an int or float, not {type(varability)}"
        return varability
 
    def node_lookback(self, node: Node) -> float:
        """Same as `lookback` but callable on a node"""
      #  if node.is_leaf:
      #      return 0
      #  else:
        return self.lookback(node.level, get_data_path(node, 'scenario'))
    
    def node_looknow(self, node: Node) -> float:
        """Same as `looknow` but callable on a node"""
     #   if node.is_leaf:
     #       return 0
     #   else:
        return self.looknow(node.level, node.data.get('eps'))
      
    @classmethod
    def from_scenario_tree(cls, scenario_tree, across_tree: bool) -> 'VariabilityProcess':
        assert scenario_tree.has_key('v'), "Some nodes do not have the data key 'v'."
        return cls(lookback_fct=lambda stage, scenario_path: inferred_lookback(stage, scenario_path, 
                                                                               scenario_tree, across_tree),
                    average_fct=lambda stage: inferred_average(stage, scenario_tree))       
    

from_scenario_tree = VariabilityProcess.from_scenario_tree


def inferred_lookback(stage, scenario_path, scenario_tree, across_tree: bool):
    if stage == 0:
        return np.std([child.data['v'] for child in scenario_tree.children]) 
    elif stage == scenario_tree.depth - 1:
        return 0
    else:
        nearest_node = list(scenario_tree.nearest_nodes(1, scenario_path, across_tree))[0]
        return np.std([child.data['v'] for child in nearest_node.children])


def inferred_average(stage, scenario_tree):
    return sum(node.data["W"] * np.std([child.data['v'] for child in node.children])
                   for node in scenario_tree.nodes_at_level(stage))
            
            
            
            


# -*- coding: utf-8 -*-

import numpy as np
from itertools import product

from .network_design_solution import NetworkDesignSolution
from ..stochastic_problem_basis import StochasticProblemBasis


class NetworkDesign(StochasticProblemBasis):
    """Two-stage Network Design Problem.
    
    Arguments:
    ----------
    param: dict with keys and values:
        'n_origins':        int
        'n_destinations':   int 
        'n_intermediates':  int
        'opening_cost':     2d-array, shape (n_vertices, n_vertices) 
        'shipping_cost':    2d-array, shape (n_vertices, n_vertices) 
        'capacity':         2d-array, shape (n_vertices, n_vertices)  
        
    where n_vertices = n_origins + n_destinations + n_intermediates.
    """
    
    def __init__(self, param):
        self.param = param        
        self.n_origins = self.param['n_origins']
        self.n_destinations = self.param['n_destinations']
        self.n_intermediates = self.param['n_intermediates']
        
        # vertices
        self.vertices = range(self.n_origins + self.n_destinations + self.n_intermediates) # all vertices
        self.origins = self.vertices[:self.n_origins] # origins
        self.destinations = self.vertices[-self.n_destinations:] # destinations
        
        # arcs
        self.od_arcs = list(product(self.origins, self.destinations)) # all arcs linking origins to destinations
        self.transport_arcs = [arc for arc in product(self.vertices, self.vertices) 
                                if arc[0] != arc[1] and arc not in self.od_arcs] # all arcs but od pairs
        self.all_arcs = self.od_arcs + self.transport_arcs


        # stochastic problem information
        StochasticProblemBasis.__init__(self, 
                                        name='Network Design Problem',
                                        n_stages=2,
                                        objective_sense='min',
                                        is_obj_random=False,
                                        is_mip=True,
                                        solution_class=NetworkDesignSolution)
                      
    # ---- random parameters ----       
    def d(self, vertex, od):
        """d_{i}^{k,s}: random demand of commodity od at vertex i in scenario s"""
        demand = self.get_rvar(1, 'd', od)
        if vertex == od[0]: # if the vertex is the origin of the commodity
            return demand
        elif vertex == od[1]: # if the vertex is the destination of the commodity
            return -demand
        else: # if the vertex is an intermediate step
            return 0
        
    # ---- 1st-stage decision variables ----    
    def y(self, arc=None):
        """y_{i,j} = 1 if arc (i,j) is open, 0 otherwise"""
        return self.get_dvar(0, 'y', arc)
    
    # ---- 2nd-stage decision variables ----    
    def z(self, arc=None, od=None):
        """z_{i,j}^{k,s}: quantity of commodity od transported on arc (i,j) in scenario s"""
        if arc is None and od is None:
            return self.get_dvar(1, 'z')
        else:
            return self.get_dvar(1, 'z', (arc, od))
 
    def decision_variables_definition(self, stage):
        """For each decision variables of type {y_{t,i} \in A : i \in I} where I is an indexing set (a list of 
        ints or tuples of ints), y_{t,i} is of type A (with A = binary (B), or integer (I), or continuous (C)), 
        and lb <= y_{t,i} <= ub, this function generates a 5-tuple of the form: ('y', I, lb, ub, A) 
        under the conditon: stage == t. (lb and ub can be two lists provided they have the same size as I)"""
        if stage == 0:
            yield 'y', self.transport_arcs, 0, 1, 'B'
        elif stage == 1:
            yield 'z', list(product(self.all_arcs, self.od_arcs)), 0, None, 'C'

    def random_variables_definition(self, stage):
        """For each random variable of type {xi_{stage,i} : i \in I} where I is an indexing set (an iterable), 
        this method generates a 2-tuple of the form ('xi', I)"""
        if stage == 1:
            yield 'd', self.od_arcs

    def objective(self):
        """Returns the problem's objective function"""        
        first_stage = self.dot(self.y(), [self.opening_cost(arc) for arc in self.transport_arcs])
        second_stage = self.dot(self.z(), [self.shipping_cost(arc) for (arc, od) in product(self.all_arcs, self.od_arcs)])
        return first_stage + second_stage
    
    def deterministic_linear_constraints(self, stage):
        """Generates all the problem's constraints that do not depend on the random parameters"""
        if stage == 1:
            yield self.transport_constraints()
                 
    def random_linear_constraints(self, stage):
        """Generates all the problem's constraints that depend on the random parameters"""
        if stage == 1: 
            yield self.demand_constraints()

    def opening_cost(self, arc):
        return self.param['opening_cost'][arc]
    
    def shipping_cost(self, arc):
        return self.param['shipping_cost'][arc]
    
    def capacity(self, arc):
        return self.param['capacity'][arc]

    def transport_constraints(self):
        # sum commodities on each each <= capacity (if arc is open)
        for arc in self.transport_arcs:
            yield self.sum([self.z(arc, od) for od in self.od_arcs]) \
                                <= self.capacity(arc) * self.y(arc), f"open_{arc}"
                                    
    def demand_constraints(self):
        # [what goes in] - [what goes out] = [demand of the node]
        for vertex, od in product(self.vertices, self.od_arcs):
            out_sum = self.sum([self.z(arc, od) for arc in self.all_arcs if arc[0] == vertex])
            in_sum = self.sum([self.z(arc, od) for arc in self.all_arcs if arc[1] == vertex])
            yield out_sum - in_sum == self.d(vertex, od), f"demand_{vertex}_{od}"
                
    def precompute_decision_variables(self, stage):
        pass
    
    def precompute_parameters(self, stage):
        pass
    
    def sanity_check(self, stage):
        pass
    
    # --- Load and save ---
    @classmethod
    def from_file(cls, path, extension='txt'):
        if extension == 'txt':
            with open(f'{path}.{extension}', "r") as f:
                file_str = f.read()
                file_str = file_str.replace('array', 'np.array')
                file_str = file_str.replace('nan', 'np.nan')
                param = eval(file_str)
        elif extension == 'pickle':
            import pickle
            with open(f'{path}.{extension}', "rb") as f:
                param = pickle.load(f)
        return cls(param)
    
    def to_file(self, path, extension='txt'):
        if extension == 'txt':
            np.set_printoptions(threshold=np.inf) # no limit on the number of elements printed in an array
            with open(f'{path}.{extension}', "w") as f:
                f.write(repr(self.param).replace("),", "),\n"))
        elif extension == 'pickle':
            import pickle
            with open(f'{path}.{extension}', "wb") as f:
                pickle.dump(self.param, f)
        else:
            TypeError(f"Extension should be 'pickle' or 'txt', not {extension}.")
                
    # --- Representation ---
    def __repr__(self):
        string_problem = StochasticProblemBasis.__repr__(self)
        string = ("Network: \n"
                  f"  {len(self.vertices)} nodes\n"      
                  f"  {len(self.transport_arcs)} transportation arcs\n"
                  f"  {len(self.od_arcs)} commodities")
        return string_problem + "\n" + string
    
from_file = NetworkDesign.from_file

def generate_random_parameters(n_origins, n_destinations, n_intermediates, seed=None):
    """Generate randomly a set of deterministic parameters of the network design problem"""
    if seed is not None:
        np.random.seed(seed)
    n_vertices = n_origins + n_destinations + n_intermediates
    # cost of opening an arc
    opening_cost = np.random.randint(3, 11, size=(n_vertices, n_vertices))
    # cost of shipping one unit on an arc
    shipping_cost = np.random.randint(5, 11, size=(n_vertices, n_vertices))
    shipping_cost[:n_origins, -n_destinations:] = 1000
    # capacity of an arc
    capacity = np.random.randint(10, 41, size=(n_vertices, n_vertices))
    capacity[:n_origins, -n_destinations:] = 10**5
    return {'n_origins': n_origins, 
            'n_destinations': n_destinations, 
            'n_intermediates': n_intermediates,
            'opening_cost': opening_cost,
            'shipping_cost': shipping_cost,
            'capacity': capacity}
    


# -*- coding: utf-8 -*-

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.patches as patches

from ..stochastic_solution_basis import StochasticSolutionBasis


class NetworkDesignSolution(StochasticSolutionBasis):
    
    def __init__(self, 
                 stochastic_problem,
                 scenario_tree):
        
        StochasticSolutionBasis.__init__(self, 
                                         stochastic_problem,
                                         scenario_tree)
         
    def is_open(self, arc):
        """Returns 1 if the arc has been open at stage 0, otherwise 0"""
        self.set_path_info()
        return self.stochastic_problem.y(arc)
    
    def x0_dict(self, forced_to_integer=False):
        return {arc: self.is_open(arc) for arc in self.stochastic_problem.transport_arcs}

    def n_open_arcs(self):
        return np.sum(self.x0['y'])
    
    def is_penalized(self, od):
        penality = 0
        for scen_index in range(self.n_scenarios):
            self.set_path_info(scen_index)
            penality += self.stochastic_problem.z(od, od)
        return True if penality > 0 else False
         
    def get_transported_amount(self, arc, scen_index):
        """Returns the amount of all commodities transported on an arc in a given scenario"""
        self.set_path_info(scen_index)
        return sum([self.stochastic_problem.z(arc, od) for od in self.stochastic_problem.od_arcs])
    
    def is_used(self, arc, scen_index):
        """True if the arc has some commodities transported on it"""
        return self.get_transported_amount(arc, scen_index) > 0
    
    def n_used_arc(self, scen_index):
        return sum([self.is_used(arc, scen_index) for arc in self.stochastic_problem.transport_arcs])
    
    def get_penalty_value(self, od, scen_index):
        self.set_path_info(scen_index)
        return self.stochastic_problem.z(od, od)
                
    def get_penalty_dict(self):
        dict_penalty_value = {od: 0 for od in self.stochastic_problem.od_arcs} # {od pair: amount transported on od arc}        
        for od in self.stochastic_problem.od_arcs:
            for scen_index in range(self.n_scenarios):
                self.set_path_info(scen_index)
                dict_penalty_value[od] += self.stochastic_problem.z(od, od) 
        return dict_penalty_value
    
    def _vertex_color(self, vertex):
        if vertex in self.stochastic_problem.origins: 
            return "C2"
        elif vertex in self.stochastic_problem.destinations: 
            return "C3"
        else: 
            return "C0"
        
    def _arc_color(self, arc):
        if arc[0] in self.stochastic_problem.origins: 
            return "C2"
        elif arc[1] in self.stochastic_problem.destinations:
            return "C3"
        elif arc[0] in self.stochastic_problem.destinations and arc[1] not in self.stochastic_problem.destinations:
            return "k"
        else:
            return "C0"
            
    def _arc_style(self, arc, scen_index=None):
        arrowstyle="Simple,tail_width=0.5,head_width=4,head_length=8" #for FancyArrowPatch
        if scen_index is None:
            if arc in self.stochastic_problem.transport_arcs:
                if self.is_open(arc): 
                    kw = dict(arrowstyle=arrowstyle, color=self._arc_color(arc), alpha=1)
                else: 
                    kw = dict(arrowstyle=arrowstyle, color="k", alpha = 0.05)
            elif arc in self.stochastic_problem.od_arcs:                
                if self.is_penalized(arc):
                    kw = dict(arrowstyle=arrowstyle, color="k", alpha = 1, linestyle = "--")
                else:
                    kw = dict(arrowstyle=arrowstyle, color="k", alpha = 0.05)    
            return kw
        else:
            if arc in self.stochastic_problem.transport_arcs:
                if self.is_used(arc, scen_index): # if open and something is transported on it
                    kw = dict(arrowstyle=arrowstyle, color=self._arc_color(arc), alpha=1)
                elif self.is_open(arc): # if open but nothing is transported on it
                    kw = dict(arrowstyle=arrowstyle, color=self._arc_color(arc), alpha=0.2)
                else: 
                    kw = dict(arrowstyle=arrowstyle, color="k", alpha = 0.05)
            elif arc in self.stochastic_problem.od_arcs:
                if self.get_penalty_value(arc, scen_index) > 0:
                    kw = dict(arrowstyle=arrowstyle, color="k", alpha = 1, linestyle = "--")
                else:
                    kw = dict(arrowstyle=arrowstyle, color="k", alpha = 0.05) 
            return kw
            
    def _vertex_label(self, vertex, scen_index):
        self.set_path_info(scen_index)
        label = ""
        if vertex in self.stochastic_problem.origins: 
            label = "{}: ".format(vertex)
            label += ",".join([str(self.stochastic_problem.d(vertex, (vertex, D)))
                                    for D in self.stochastic_problem.destinations])
        elif vertex in self.stochastic_problem.destinations: 
            label = "{}: ".format(vertex)
            label += ",".join([str(self.stochastic_problem.d(vertex, (O, vertex)))
                                    for O in self.stochastic_problem.origins])
        return label
                
    def plot_network(self, 
                     scen_index=None, 
                     show_title=True, 
                     randomize_txt_box=True,
                     figsize=(6,5), 
                     path=None, 
                     extension=".pdf"):
        fig, ax = plt.subplots(figsize=figsize)
        N = len(self.stochastic_problem.vertices)
        pos = {vertex: np.exp(2*np.pi*1j * k / N) for k, vertex in enumerate(self.stochastic_problem.vertices, 1)}
        pos = {key: (value.real, value.imag) for key, value in pos.items()}
        
        # Plot vertices
        for vertex in self.stochastic_problem.vertices:
            label = self._vertex_label(vertex, scen_index) if scen_index is not None else None
            ax.scatter(*pos[vertex], c=self._vertex_color(vertex), 
                       s=150, alpha=0.8, label=label)
            ax.text(*pos[vertex], vertex, 
                    horizontalalignment='center', verticalalignment='center')   
            
        # Plot arcs
        for arc in self.stochastic_problem.all_arcs:             
            a = patches.FancyArrowPatch(pos[arc[0]], pos[arc[1]], 
                                        connectionstyle="arc3,rad=.05", 
                                        **self._arc_style(arc, scen_index))
            fig.gca().add_patch(a)
            if scen_index is not None:
                self._print_transportated_value(arc, scen_index, randomize_txt_box, pos, ax)
                
        if scen_index is not None:
            ax.legend(bbox_to_anchor=(1, 1.15), loc=1, borderaxespad=1, prop={'size': 7})
        if show_title:
            if scen_index is None:
                obj_to_show = self.objective_value
                ax.set_title(f"$v^*: {obj_to_show:.2f}$")
            else:
                obj_to_show = self.objective_value_at_leaves[scen_index]
                ax.set_title(f"$v^*: {obj_to_show:.2f}$")
        ax.axis('off')
        self._save_fig(path, extension)
        fig.show()
           
    def _print_transportated_value(self, arc, scen_index, randomize_txt_box, pos, ax):  
        if not hasattr(self, 'alpha'):
            # coef of the convex combination giving the position of the box between the two vertices
            self.alpha = {arc: 0.1 for arc in self.stochastic_problem.all_arcs} 
            
        if self.is_used(arc, scen_index): # only arcs with commodity transported on it have a txt box
            # color of the box
            if arc[0] in self.stochastic_problem.origins:
                box_color = 'C2' 
            elif arc[1] in self.stochastic_problem.destinations:
                box_color = 'C3'
            else:
                box_color = 'C0'
            props = dict(boxstyle='round', facecolor=box_color, alpha=0.3)
            # location of the box
            if randomize_txt_box:
                n_adjacent_arcs = sum([self.is_used(a, scen_index) for a in self.stochastic_problem.all_arcs 
                                           if a[0] == arc[0] and a != arc])
                self.alpha[arc] = 0.1 if n_adjacent_arcs == 0 else np.random.uniform(0.05, 0.2)
            x, y = (1 - self.alpha[arc]) * np.array(pos[arc[0]]) + self.alpha[arc] * np.array(pos[arc[1]])
            # plot text in box
            txt_to_print = int(round(self.get_transported_amount(arc, scen_index), 0))
            ax.text(x, y, txt_to_print, fontsize=10, 
                    horizontalalignment='center', verticalalignment='center', bbox=props)
                    
    def _save_fig(self, path, extension):
        if path:
            plt.savefig(path + extension)
            print(f"Figure saved at {path + extension}") 
            
    @classmethod
    def from_file(cls, 
                  path_tree, 
                  path_problem, 
                  extension_tree='pickle', 
                  extension_prob='txt'):
        """ Load network design solution from two files: one for the problem and one for the scenario tree.
        Arguments:
        ----------
        path_tree: string
        extension_tree: {'pickle', 'txt'}, see: Node.from_file().
        path_problem: string
        extension_prob: {'pickle', 'txt'}, see: MinePlanningProblem.from_file()             
        """            
        from stochoptim.stochprob.network_design.network_design_problem import NetworkDesign
        from stochoptim.scengen.scenario_tree import ScenarioTree
        problem = NetworkDesign.from_file(path_problem, extension_prob)
        scenario_tree = ScenarioTree.from_file(path_tree, extension_tree)
        return cls(problem, scenario_tree)

from_file = NetworkDesignSolution.from_file


# -*- coding: utf-8 -*-

from abc import ABC, abstractmethod
from typing import Dict

from itertools import product
import numpy as np

from stochoptim.scengen.scenario_tree import ScenarioTree

class UncertaintyBasis(ABC):
    """
    Basis for the uncertainty of two-stage stochastic problems from a given set of scenarios.
    """
    
    def __init__(self, name: str, n_scenarios: int, features: Dict[str, int]):
        self.name = name
        self.n_scenarios = n_scenarios
        self.features = features
        self.n_features = sum(self.features.values())
            
    @abstractmethod
    def get_scenario(self, scen_index) -> Dict[str, np.ndarray]:
        """Returns a set of scenarios.
        
        Arguments:
        ----------
        scen_index: int
            The index of the scenario to return.
            
        Returns:
        --------
        Dict[str, 1d-array]: mapping a variable name to a variable array.
        """  
        pass
    
    def get_scenario_tree(self, scen_indices=None):
        """Returns a scenario tree.
        
        Arguments:
        ----------
        scen_indices: list of ints or None (default: None)
            The indices of the scenarios returned. If None, all scenarios are returned.
            
        Returns:
        --------
        instance of ScenarioTree
        """   
        if scen_indices is None:
            scen_indices = range(self.n_scenarios)
        weights = np.ones(len(scen_indices)) / len(scen_indices)        
        data_dict = {(): {'W': 1}}
        for i, scen_index in enumerate(scen_indices):
            data_dict[(i,)] =  {'W': weights[i], 'scenario': self.get_scenario(scen_index)}            
        return ScenarioTree.from_data_dict(data_dict)
        
    def _check_scen_index(self, scen_index):
        assert scen_index in range(self.n_scenarios), \
            f"Scenario index {scen_index} is not in the range [0, {self.n_scenarios-1}]."
            
    def __repr__(self):
        string = (f"{self.name} \n"
                  f"  - scenarios: {self.n_scenarios} \n"
                  f"  - features: {self.n_features} \n")
        for var_name, n_var in self.features.items():
            string += f"    - {var_name}: {n_var} \n"
        return string
    
    
class CrossUncertainty(UncertaintyBasis):
    
    def __init__(self, *uncertainties, combiner):
        self.uncertainties = uncertainties
        self.combiner = combiner
        self.n_uncertainties = len(self.uncertainties)
        
        self.n_scenarios_per_uncertainty = [uncertainty.n_scenarios 
                                            for uncertainty in self.uncertainties]
        self.n_features_per_uncertainty = [uncertainty.n_features 
                                           for uncertainty in self.uncertainties]
        self._set_combiner()
            
        # combine all features together
        features = {}
        for uncertainty in self.uncertainties:
            features = {**features, **uncertainty.features}
            
        UncertaintyBasis.__init__(self, 
                                  name=f"Cross uncertainty ({self.combiner})",
                                  n_scenarios=self.n_scenarios,
                                  features=features)
        
    def _set_combiner(self):
        if self.combiner == "cartesian": 
            self.n_scenarios = np.prod(self.n_scenarios_per_uncertainty)
            self.map_scen_index_to_scen_tuple = list(
                    product(*[range(uncertainty.n_scenarios) 
                    for uncertainty in self.uncertainties])
            )
        elif self.combiner == "zip":
            assert len(set(self.n_scenarios_per_uncertainty)) == 1, \
                "All uncertainties must have the same number of scenarios for combiner 'zip'."
            self.n_scenarios = self.n_scenarios_per_uncertainty[0]
            self.map_scen_index_to_scen_tuple = [tuple([i] * len(self.uncertainties)) 
                                                    for i in range(self.n_scenarios)]
        else:
            raise TypeError(f"Wrong 'combiner' input: must be either 'cartesian' or 'zip', "
                            "not {self.combiner}")    
          
    def get_scenario(self, scen_index):
        self._check_scen_index(scen_index)
        scen_tuple = self.map_scen_index_to_scen_tuple[scen_index]
        scenario = {}
        for i, uncertainty in enumerate(self.uncertainties):
            scenario = {**scenario, **uncertainty.get_scenario(scen_tuple[i])}
        return scenario
            
    def _rows_sub_cross_uncertainty(self, uncertainty_indices):
        """Returns the rows (scenarios) indices that are required to build a sub cross-uncertainty.
        
        Arguments:
        ----------
        uncertainty_indices: list of ints
            The indices of the uncertainties used to build the sub cross-uncertainty.
            
        Returns:
        --------
        list of ints
        """
        rows = np.linspace(0, self.n_scenarios-1, self.n_scenarios, dtype='int32')
        rows = rows.reshape(*self.n_scenarios_per_uncertainty)
        iterator = product(*[range(n_scenarios) if index in uncertainty_indices else range(1) 
                            for index, n_scenarios in enumerate(self.n_scenarios_per_uncertainty)])
        return [rows[tuple_index] for tuple_index in iterator]
                
    def __repr__(self):
        string = UncertaintyBasis.__repr__(self) + "\nComposed of:\n------------"
        for uncertainty in self.uncertainties:
            string += "\n" + UncertaintyBasis.__repr__(uncertainty)
        return string


# -*- coding: utf-8 -*-

from __future__ import division
from __future__ import unicode_literals

from typing import Dict, Union, Optional, List
import copy
import numpy as np
import matplotlib.pyplot as plt
from scipy import sparse

from stochoptim.scengen.tree_structure import Node
from stochoptim.scengen.tree_structure import get_data_path
from stochoptim.scengen.scenario_process import ScenarioProcess
from stochoptim.scengen.figure_of_demerit import FigureOfDemerit
from stochoptim.scengen.variability_process import VariabilityProcess


class ScenarioTree(Node):

    def __init__(self, node: Node):
        """
        Arguments:
        ----------
        node: Node
            Root of the scenario tree. 
            If it is not already a root (i.e., if it has a parent), it is made one by cutting its parent. 
        """
        node.make_it_root()
        super().__init__(*node.children, **node.data)
        
    # --- Properties ---   
    def are_scenarios_consistent(self):
        """Return True if all the paths have the same number of variables."""
        array_map_stage_to_rvar_nb = np.array([self._map_stage_to_rvar_nb(leaf) for leaf in self.leaves])
        return (array_map_stage_to_rvar_nb == array_map_stage_to_rvar_nb[0]).all()
    
    @property
    def map_stage_to_rvar_names(self) -> Dict[int, List[str]]:
        return {stage: list(var_dict.keys()) for stage, var_dict in self.map_stage_to_rvar_nb.items()}
    
    @property
    def map_stage_to_rvar_nb(self) -> Dict[int, Dict[str, int]]:
        bottom_most_leaf = self.node_at_address((0,) * (self.depth - 1))
        return self._map_stage_to_rvar_nb(bottom_most_leaf)
    
    def _map_stage_to_rvar_nb(self, leaf: Node) -> Dict[int, Dict[str, int]]:
        """Return the variables names and numbers at each level along a certain path leading to a leaf."""
        return {node.level: {var_name: len(variables) for var_name, variables in node.data['scenario'].items()}
                                 for node in leaf.branch if isinstance(node.data.get('scenario'), dict)}
        
    @staticmethod
    def get_scenario_path(node: Node):
        return get_data_path(node, 'scenario')
        
    def get_figure_of_demerit(self, demerit: Optional[FigureOfDemerit]):
        """Figure of demerit of the scenario tree, as given by the exact formula.
        
        Arugment:
        ---------
        demerit: FigureOfDemerit
            The figure of demerit used to compute the demerit of the children at each node.
            
        Returns:
        --------
        float > 0
        """
        self._is_filled()
        return demerit(self, subtree=True, path=True)
    
    # --- Scenarios filling ---
    def fill(self, 
             scenario_process: ScenarioProcess, 
             optimized: Optional[str] = None, 
             variability_process: Optional[VariabilityProcess] = None, 
             demerit: Optional[FigureOfDemerit] = None):
        """Fills the tree structure with scenarios given by the scenario process.
        
        Arguments:
        ----------           
        optimized: {'forward', 'backward', None}
            The way the scenario tree is filled with scenarios and weights.
            If 'forward', the assignment of scenarios to nodes is optimized from the root to 
            the leaves, and it is guided by the scenario values (under data key 'scenario').
            If 'backward', the assignment of scenarios to nodes is optimized from the leaves 
            to the root, and it is guided by the epsilon values (under data key 'eps').
            If None, the assignment is not optimized.
        
        scenario_process: ScenarioProcess
            The scenario process used to generate the scenarios.

        variability_process: VariabilityProcess
            The variability process used to guide the assignement of scenarios to nodes.

        demerit: FigureOfDemerit
            The figure of demerit used to compute the demerit of the children at each node.
        """        
        self._fill_epsilon(scenario_process) # assign weights ('W') and optionaly epsilon sample ('eps') at each node
        if optimized is None:
            self._fill_scenario(scenario_process)
        elif optimized == 'forward':
            assert variability_process.has_lookback(), \
                "The variability process must have a `lookback` method for the forward-optimized scenario tree."
            self._optimized_assignment_forward(scenario_process, variability_process, demerit) 
        elif optimized == 'backward':
            assert variability_process.has_looknow(), \
                "The variability process must have a `looknow` method for the backward-optimized scenario tree."
            self._optimized_assignment_backward(variability_process, demerit) 
            self._fill_scenario(scenario_process)
        else:
            raise ValueError(f"Wrong 'optimized' keyword: must be None, 'forward', or 'backward', not {optimized}.")

    def _fill_epsilon(self, scenario_process):
        """Fills the scenario tree with points 'eps' and weights 'w'"""
        self.data["w"] = 1 # root
        for node in self.nodes:
            if not node.is_leaf:
                weights, epsilons = scenario_process.get_children_sample(node)
                random_indices = np.random.permutation(range(len(node.children)))
                for i, child in zip(random_indices, node.children):
                    child.data["w"] = weights[i]
                    if epsilons is not None:
                        child.data["eps"] = epsilons[i]

    def _fill_scenario(self, scenario_process):
        """Fills the scenario tree with points 'scenario' and weights 'W' given the assignment of 'eps' 
        and 'w' at each node."""
        for node in self.nodes:
            node.data["W"] = node.parent.data["W"] * node.data["w"] if not node.is_root else 1
            root_scenario = scenario_process.get_node_scenario(node, path=False)
            if root_scenario is not None:
                node.data["scenario"] = root_scenario
            
    def _optimized_assignment_forward(self, scenario_process, variability_process, demerit):  
        for node in self.forward_nodes:
            if node.is_root:
                node.data["W"] = 1
                root_scenario = scenario_process.get_node_scenario(node, path=False)
                if root_scenario is not None:
                    node.data["scenario"] = root_scenario
                
            if not node.is_leaf and not node.is_parent_of_leaf: 
                # sort child (in place) by decreasing demerit
                node.children.sort(key=lambda c: -demerit(c, subtree=False)) 
                # sort child (not in place) by increasing variability
                for child in node.children:
                    child.data["scenario"] = scenario_process.get_node_scenario(child, path=False)
                child_by_variability = sorted(node.children, 
                                              key=lambda c: variability_process.node_lookback(c) * c.data["w"])
                weights = [c.data["w"] for c in child_by_variability]
                epsilons = [c.data.get("eps") for c in child_by_variability]
                scenarios = [c.data['scenario'] for c in child_by_variability]
                
                for i, child in enumerate(node.children):
                    child.data["w"] = weights[i]
                    child.data["W"] = node.data["W"] * child.data["w"]
                    if epsilons[i] is not None:
                        child.data["eps"] = epsilons[i]
                    child.data["scenario"] = scenarios[i]
                    
            if node.is_parent_of_leaf:
                for child in node.children:
                    child.data["scenario"] = scenario_process.get_node_scenario(child, path=False)
                    child.data["W"] = node.data["W"] * child.data["w"]
                    
    def _optimized_assignment_backward(self, variability_process, demerit):
        for node in self.backward_nodes:                
            if not node.is_leaf and not node.is_parent_of_leaf: 
                # sort child (in place) by decreasing demerit
                node.children.sort(key=lambda c: -demerit(c, subtree=True, path=False)) 
                # sort child (not in place) by increasing variability
                child_by_variability = sorted(node.children, 
                                              key=lambda c: variability_process.node_looknow(c) * c.data["w"])
                weights = [c.data["w"] for c in child_by_variability]
                epsilons = [c.data.get("eps") for c in child_by_variability]
                
                for i, child in enumerate(node.children):
                    child.data["w"] = weights[i]
                    if epsilons[i] is not None:
                        child.data["eps"] = epsilons[i]
    
    def _is_filled(self):
        """Check whether the tree is filled with scenarios."""
        assert self.has_key('scenario', excluding_levels=[0]), \
            "Fill the tree structure before computing the scenario-tree figure of demerit."
     
    # --- Forward generation ---                 
    @classmethod
    def forward_generation(cls, 
                           n_stages: int,
                           n_scenarios: int, 
                           scenario_process: ScenarioProcess, 
                           variability_process: VariabilityProcess, 
                           alpha: float):
        """Generates a scenario tree by the forward bound-minimizing heuristic.
        
        Arguments:
        ----------
        n_stages: int >= 1
        
        n_scenarios: int >= 1
                
        scenario_process: ScenarioProcess
            The scenario process used to generate the scenarios.

        variability_process: VariabilityProcess
            The variability process used to guide the assignement of scenarios to nodes.
            
        alpha: float > 0
            Convergence rate of the discretization method (typically from 0.5 to 2).
            
        Returns:
        --------
        ScenarioTree
        """
        assert variability_process.has_average(), \
            "The variability process must have an `average_fct` method for the forward-generation algorithm."
        avg_variability = [variability_process.average(stage) for stage in range(n_stages-1)]
        width_vector = cls.optimal_width_vector(n_scenarios, alpha, avg_variability)
        return cls.forward_generation_from_given_width(width_vector, scenario_process, variability_process, alpha)
         
    @classmethod
    def forward_generation_from_given_width(cls, 
                                            width_vector: List[int], 
                                            scenario_process: ScenarioProcess, 
                                            variability_process: VariabilityProcess, 
                                            alpha: float):
        """Generates a scenario tree by the forward bound-minimizing heuristic.
        
        Arguments:
        ----------
        width_vector: list of int >= 1
        
        scenario_process: ScenarioProcess
            The scenario process used to generate the scenarios.

        variability_process: VariabilityProcess
            The variability process used to guide the assignement of scenarios to nodes.
            
        alpha: float > 0
            Convergence rate of the discretization method (typically from 0.5 to 2).
                
        Returns:
        --------
        ScenarioTree
        """
        last_stage = len(width_vector)
        tree = cls.from_data_dict({(): {"M": width_vector[0], "W": 1, "w": 1}})       
        root_scenario = scenario_process.get_node_scenario(tree, path=False)
        if root_scenario is not None:
            tree.data["scenario"] = root_scenario             
        tree.data["g"] = variability_process.node_lookback(tree)
        
        # difference between the actual width of the leaves and the target one
        node_gap = lambda tree: sum(leaf.data["M"] for leaf in tree.leaves) - width_vector[tree.depth-1]
        
        for stage in range(1, last_stage):
            # 1. Extend and fill the structure
            tree._extend_tree_by_one_stage(scenario_process)
            for leaf in tree.leaves:
                leaf.data["g"] = variability_process.node_lookback(leaf)
            # 2. Compute the optimal number of child nodes
            if width_vector[stage] == width_vector[stage-1]: 
                for leaf in tree.leaves:
                    leaf.data["M"] = 1
            else:
                normalization = sum((leaf.data["W"] * leaf.data["g"])**(1/(alpha+1)) for leaf in tree.leaves)
                for leaf in tree.leaves:
                    leaf.data["m"] = (width_vector[stage] / normalization) \
                                        * (leaf.data["W"] * leaf.data["g"])**(1/(alpha+1))
                    leaf.data["M"] = int(max(1, round(leaf.data["m"])))

            # 3. Correct the rounding off of the number of child nodes (if necessary) so that the actual width 
            # equal the target
            while node_gap(tree) > 0:
                leaf = min([leaf for leaf in tree.leaves if leaf.data["M"] >= 2], 
                               key = lambda leaf: abs(leaf.data["m"] - (leaf.data["M"] - 1)))
                leaf.data["M"] = leaf.data["M"] - 1 # remove one child
                
            while node_gap(tree) < 0:
                leaf = min(tree.leaves, 
                           key = lambda leaf: abs(leaf.data["m"] - (leaf.data["M"] + 1)))
                leaf.data["M"] = leaf.data["M"] + 1 # add one child
                
        # extend and fill the last stage
        tree._extend_tree_by_one_stage(scenario_process)
        # delete temporary data
        tree.delete_data(["M", "m", "g"])
        assert tree.width == list(width_vector),  ("Missmatch between the actual tree width and the target one "
                                            f"actual width: {tree.width}, target width: {list(width_vector)}")
        return tree

    @staticmethod
    def optimal_width_vector(n_scenarios: int, 
                             alpha: float, 
                             gamma: List[float]):
        bush = [1 if gamma[m] == 0 else None for m in range(len(gamma))] # branching factor = 1 if no variability
        inactive_set = [m for m in range(len(gamma)) if gamma[m] != 0] # set of inactive constraints
        found = False
        while not found:
            found = True
            denominator = np.prod([gamma[i]**(1/alpha) for i in inactive_set])**(1/len(inactive_set))
            for m in inactive_set:
                bush[m] = n_scenarios**(1/len(inactive_set)) * gamma[m]**(1/alpha) / denominator
                found = found and (bush[m] > 1)
            if not found:
                # find the index m* such that gamma[m] is the smallest
                min_index = min(inactive_set, key = lambda m: gamma[m]) 
                # remove m* from the set of inactive constraints
                inactive_set.remove(min_index)
                bush[min_index] = 1
        width_vector = np.round(np.cumprod(bush)).astype('int')
        return list(width_vector)
    
    def _extend_tree_by_one_stage(self, scenario_process):
        for leaf in list(self.leaves):
            leaf.add(*[Node() for i in range(leaf.data["M"])])
            weights, epsilons = scenario_process.get_children_sample(leaf)
            for i, child in enumerate(leaf.children):
                child.data["w"] = weights[i]
                if epsilons is not None:
                    child.data["eps"] = epsilons[i]
                child.data["W"] = leaf.data["W"] * child.data["w"]
                child.data["scenario"] = scenario_process.get_node_scenario(child, path=False)
                
    # --- Alternative constructors ---
    @staticmethod
    def _set_equal_weights(tree_structure):
        tree_structure.data["W"] = 1
        for node in tree_structure.nodes:
            if not node.is_root:
                node.data["W"] = node.parent.data["W"] / len(node.parent.children)
        return tree_structure
    
    @classmethod
    def from_topology(cls, topology, equal_weights=False):
        scen_tree = cls(Node.from_topology(topology))
        if equal_weights:
            scen_tree = ScenarioTree._set_equal_weights(scen_tree)
        return scen_tree
    
    @classmethod
    def from_recurrence(cls, last_stage, init, recurrence, equal_weights=False):
        scen_tree = cls(Node.from_recurrence(last_stage, init, recurrence))
        if equal_weights:
            scen_tree = ScenarioTree._set_equal_weights(scen_tree)
        return scen_tree
    
    @classmethod
    def from_bushiness(cls, bushiness, equal_weights=False):
        scen_tree = cls(Node.from_bushiness(bushiness))
        if equal_weights:
            scen_tree = ScenarioTree._set_equal_weights(scen_tree)
        return scen_tree
    
    @classmethod
    def from_data_dict(cls, data_dict, equal_weights=False):
        scen_tree = cls(Node.from_data_dict(data_dict))
        if equal_weights:
            scen_tree = ScenarioTree._set_equal_weights(scen_tree)
        return scen_tree

    @classmethod
    def twostage_from_scenarios(cls, 
                                scenarios, 
                                n_rvar: Dict[str, int],
                                weights=None):
        """
        Constructor of a two-stage scenario tree directly from the set of scenarios and the weights.
    
        Arguments:
        ----------
        scenarios: 2d-array-like
            Array of shape (number_of_scenarios, dimension_of_a_scenario).
    
        n_rvar: Dict[str, int] (optional)
            Dictionary mapping each random variable's name to the number of such variables.
            If None, one variable with name "" is created.
            
        weights: 1d-array (optional)
            Array of shape (number_of_scenarios,). If None, equal-weights are considered.
        """
        if sparse.issparse(scenarios):
            if scenarios.dtype == np.float16: # float16 not directly supported for sparse matrix
                scenarios = scenarios.astype(np.float32).toarray().astype(np.float16)
            else:
                scenarios = scenarios.toarray()
        if isinstance(scenarios, list):
            scenarios = np.array(scenarios)
        assert len(scenarios.shape) == 2, \
            f"The scenarios must be given as a 2d-array, not a {len(scenarios.shape)}d-array."
        n_scenarios, dim_scenario = scenarios.shape
        if weights is None:
            weights = np.ones(n_scenarios) / n_scenarios
            
        if n_rvar is None:
            n_rvar = {"": dim_scenario}
        else:
            assert sum(n_rvar.values()) == dim_scenario, ("Mismatch between the number of random variables "
            f"in `n_rvar` ({sum(n_rvar.values())}) and the number of features in the scenarios ({dim_scenario})")
            
        data_dict = {(): {'W': 1}}
        for i in range(n_scenarios):
            split_points = np.cumsum(np.array(list(n_rvar.values())))
            split_scenario = np.split(scenarios[i], split_points)
            data_dict[(i,)] =  {'scenario': {var_name: split_scenario[j] 
                                                for j, var_name in enumerate(n_rvar.keys())}, 
                                'W': weights[i]}
    
        return cls.from_data_dict(data_dict)
    
    @classmethod
    def combtree_from_scenarios(cls, 
                                scenarios: np.ndarray,
                                map_stage_to_rvar_nb: Dict[int, Dict[str, int]],
                                weights: Optional[np.ndarray] = None):
        """
        Constructor of a multi-stage scenario tree with comb structure from a set of scenarios and their weights.
        A comb structure has all its scenarios linked at the root only.
        
        Arguments:
        ----------
        scenarios: 2d-array of shape (number_of_scenarios, number_of_features)
            The features should be ordered by stage and within a stage by the order in the list of variable names.
            
        map_stage_to_rvar_nb: Dict[int, Dict[str, int]]
            Map stage (int) to a map between the variable names (str) and the variables numbers (int) at that stage.
        
        weights: 1d-array (optional)
            Array of shape (number_of_scenarios,). If not provided, equal-weights are considered.
        """
        if sparse.issparse(scenarios):
            if scenarios.dtype == np.float16: # float16 not directly supported for sparse matrix
                scenarios = scenarios.astype(np.float32).toarray().astype(np.float16)
            else:
                scenarios = scenarios.toarray()
        last_stage = max(map_stage_to_rvar_nb.keys())
        n_var_at_each_stage = [sum(map_stage_to_rvar_nb.get(t, {'': 0}).values()) for t in range(1, last_stage+1)]
        
        # check whether each stage has at least one random variable
        assert (np.array(n_var_at_each_stage) >= 1).all(), f"One stage has no random variable: {map_stage_to_rvar_nb}"
        # check whether the number of variables matches between the input scenarios and map_stage_to_rvar_nb
        assert sum(n_var_at_each_stage) == scenarios.shape[1], \
        (f"Mismatch between the number of random variables expected from `map_stage_to_rvar_nb` "
         f"({sum(n_var_at_each_stage)}) and the number of features in `scenarios` ({scenarios.shape[1]}).")
        
        if weights is None:
            n_scenarios = scenarios.shape[0]
            weights = np.ones(n_scenarios) / n_scenarios

        data_dict = {(): {'W': 1}}
        for i, path_scenario in enumerate(scenarios):
            # decompose into a list of scenario at each stage
            split_points = np.cumsum([sum(map_stage_to_rvar_nb[t].values()) for t in map_stage_to_rvar_nb.keys()])[:-1]
            split_path_scenario = np.split(path_scenario, split_points)
            
            for t, stage_scenario in enumerate(split_path_scenario, 1):
                # decompose into a list of scenario for each var_name
                split_points = np.cumsum(np.array(list(map_stage_to_rvar_nb[t].values())))
                split_stage_scenario = np.split(stage_scenario, split_points)
                # append to data dict
                address = (i,) + tuple(0 for _ in range(t-1))
                data_dict[address] =  {'scenario': {var_name: split_stage_scenario[j] 
                                                        for j, var_name in enumerate(map_stage_to_rvar_nb[t].keys())}, 
                                        'W': weights[i]}
    
        return cls.from_data_dict(data_dict)

    # --- Operations on scenarios ---   
    def merge(self, **kwargs):
        """Merge sibling nodes if they have identical scenarios. Adjust the weights 
        accordingly. (Note that the nodes are merged regardless of whether they have identical
        subtrees or not.)
        
        kwargs: 
        ------
        All kwargs of np.isclose
        """
        for node in self.nodes:
            if node.is_leaf:
                continue
            to_be_removed = []
            for k, child1 in enumerate(node.children):
                for child2 in node.children[k+1:]:
                    # test if same scenario at child node
                    is_equal = True
                    for var_name in self.map_stage_to_rvar_names[node.level + 1]:
                        if not np.isclose(child1.data['scenario'][var_name], 
                                          child2.data['scenario'][var_name], 
                                          **kwargs).all():
                            is_equal = False
                            break
                    if is_equal:
                        weight_coef = (child2.data["W"] + child1.data["W"]) / child2.data["W"]
                        for n in child2.nodes:
                            n.data["W"] *= weight_coef
                        to_be_removed.append(child1)
                        break
            for child in to_be_removed:
                child.remove()
                    
    def average(self, 
                map_stage_to_rvar_names: Optional[Dict[int, List[str]]] = None,
                across_tree: bool = True):
        """Replace some scenarios by their average value in place.        
        
        Arguments:
        ---------
        across_tree: bool (default: True)
            If True, averages are computed across all nodes at a given stage. Otherwise, it
            is compute across all children.
            
        map_stage_to_rvar_names: Dict[int, List[str]] or None (default: None)
            The stages (int) and variables names (List[str]) for which the scenarios are averaged.
            If None, all stages and all variables are averaged.
        """
        if across_tree:
            self._average_across_tree(map_stage_to_rvar_names)
        else:
            self._average_across_children(map_stage_to_rvar_names)

    def _average_across_tree(self, map_stage_to_rvar_names: Optional[Dict[int, List[str]]] = None):
        """Replace some scenarios by their average value in place.
        
        Argument:
        ---------
        map_stage_to_rvar_names: Dict[int, List[str]] or None (default: None)
            The stages (int) and variables names (List[str]) for which the scenarios are averaged.
            If None, all stages and all variables are averaged.
        """
        if map_stage_to_rvar_names is None:
            map_stage_to_rvar_names = self.map_stage_to_rvar_names
        for stage in map_stage_to_rvar_names.keys():
            for var_name in map_stage_to_rvar_names[stage]:
                avg_scen = np.mean(self.to_numpy({stage: [var_name]}), axis=0)
                for node in self.nodes_at_level(stage):
                    node.data['scenario'][var_name] = avg_scen

    def _average_across_children(self, map_stage_to_rvar_names: Optional[Dict[int, List[str]]] = None):
        """Replace some scenarios by their average value in place.
        
        Argument:
        ---------
        map_stage_to_rvar_names: Dict[int, List[str]] or None (default: None)
            The stages (int) and variables names (List[str]) for which the scenarios are averaged.
            If None, all stages and all variables are averaged.
        """
        if map_stage_to_rvar_names is None:
            map_stage_to_rvar_names = self.map_stage_to_rvar_names
        for stage in map_stage_to_rvar_names.keys():
            for node in self.nodes_at_level(stage - 1):
                for var_name in map_stage_to_rvar_names[stage]:                
                    avg_scen = np.mean([child.data['scenario'][var_name] for child in node.children], axis=0)
                    for child in node.children:
                        child.data['scenario'][var_name] = avg_scen
            
    def to_numpy(self, map_stage_to_rvar_names: Optional[Dict[int, List[str]]] = None) -> np.ndarray:
        """ Return the scenarios as a numpy array. 
        
        Scenarios on the same path but at different stages, or corresponding to different random variables, 
        are concatenated along axis = 1 by stage and within a stage by the order of the variable names in input list. 
        Scenarios on different paths are placed along axis = 0.
        
        Argument:
        ---------
        map_stage_to_rvar_names: Dict[int, List[str]] or None (default: None)
            The stages (int) and variables names (List[str]) for which the scenarios are put in an array.
            If None, it returns the scenarios at all stages and for all variable names.
            
        Returns:
        --------
        2d-array of shape (n_leaves, N) where n_leaves is the number of leaves in the scenario tree and
            N is the number of random variables.
        """
        if map_stage_to_rvar_names is None:
            map_stage_to_rvar_names = self.map_stage_to_rvar_names
        return self.get_subtree_as_numpy(self, map_stage_to_rvar_names)
        
    def get_path_as_numpy(self, node, map_stage_to_rvar_names: Optional[Dict[int, List[str]]] = None) -> np.ndarray:
        """Return the scenarios as a numpy 1d-array along the path leading to a certain node. 
        Scenarios are concatenated by stage and within a stage by the order of variables in the input list."""
        if map_stage_to_rvar_names is None:
            map_stage_to_rvar_names = self.map_stage_to_rvar_names
        return np.concatenate([m.data['scenario'][var_name] for m in node.branch if not m.is_root
                                                               if m.level in map_stage_to_rvar_names.keys()
                                                                   for var_name in map_stage_to_rvar_names[m.level]])
    
    def get_subtree_as_numpy(self, node, map_stage_to_rvar_names: Optional[Dict[int, List[str]]] = None) -> np.ndarray:
        """Return the scenarios in a subtree as a numpy array (excluding the scenarios at the subtree root).
        Scenarios on the same path but at different stages, or corresponding to different random variables, are 
        concatenated along axis = 1 by stage and within a stage by the order of the variable names in input list. 
        Scenarios on different paths are placed along axis = 0.
        
        Argument:
        ---------
        map_stage_to_rvar_names: Dict[int, List[str]] or None (default: None)
            The stages (int) and variables names (List[str]) for which the scenarios are put in an array.
            If None, it returns the scenarios at all stages and for all variable names.
            
        Returns:
        --------
        2d-array of shape (n_leaves, N) where n_leaves is the number of leaves in the subtree rooted at `node` and 
            N is the number of random variables.
        """
        if map_stage_to_rvar_names is None:
            map_stage_to_rvar_names = self.map_stage_to_rvar_names
        return np.array([self.get_path_as_numpy(n, map_stage_to_rvar_names) for n in node.leaves if n != node])
    
    # --- Plots ---
    def plot_scenarios(self, 
             var_name: Optional[Union[str, Dict[int, str]]] = None, 
             scenario_precision: int = 2,
             format_weights: str = '.3f',
             **kwargs):
        if var_name is None:
            print_on_nodes = None
        elif isinstance(var_name, str):
            def print_on_nodes(node):
                if node.data.get('scenario') is None:
                    return ""
                elif node.data['scenario'].get(var_name) is None:
                    return ""
                else:
                    return np.array_str(node.data.get('scenario').get(var_name), 
                                        precision=scenario_precision)
        else:
            def print_on_nodes(node):
                if node.data.get('scenario') is None:
                    return ""
                elif var_name.get(node.level) is None:
                    return ""
                elif node.data['scenario'].get(var_name[node.level]) is None:
                    return ""
                else:
                    return np.array_str(node.data['scenario'][var_name[node.level]], 
                                        precision=scenario_precision)
            
        def print_on_edges(node):
            if node.data.get('W') is not None:
                return f"{node.data.get('W'):{format_weights}}"
            else:
                return ""

        return super().plot(print_on_nodes=print_on_nodes,
                             print_on_edges=print_on_edges,
                             **kwargs)
    
    def plot_trajectories(self, 
                       var_name,
                       component=0,
                       figsize=(10,5),
                       color=None):                
        fig, ax = plt.subplots(figsize=figsize)
        i = 0
        color_fct = lambda i: f"C{i%10}" if color is None else lambda i: color
        for node in self.nodes:
            # plot dots at all nodes but leaves
            if not node.is_leaf:
                if node.data.get("scenario", {}).get(var_name) is None:
                    continue
                plt.scatter(node.level, node.data["scenario"][var_name][component], marker="", c=color_fct(i))
                #plot links between the dots
                for m in node.children:
                    if m.data.get("scenario", {}).get(var_name) is None:
                        continue
                    x = [node.level, m.level]
                    y = [node.data["scenario"][var_name][component], m.data["scenario"][var_name][component]]
                    ax.plot(x, y, c=color_fct(i))
                    # plot dots at leaves
                    if node.is_parent_of_leaf:
                        ax.scatter(m.level, m.data["scenario"][var_name][component], marker="", c=color_fct(i))
                i += 1
        return ax
        
    def plot_hist(self, 
                  stage,
                  var_name,
                  component=0,
                  bins=10,
                  figsize=(5,5),
                  return_mean_std=False,
                  xlim=None,
                  ax=None):
        """Plot the weighted histogram of the scenario-tree values at stage."""
        assert stage <= self.depth-1, f"Stage {stage} is higher than maximum scenario-tree stage {self.depth-1}."
        if ax is None:
            fig, ax = plt.subplots(figsize=figsize)
        hist_data = [node.data["scenario"][var_name][component] for node in self.nodes_at_level(stage)]
        hist_weight = [node.data["W"] for node in self.nodes_at_level(stage)]
        ax.hist(hist_data, bins=bins, density=True, weights=hist_weight)
        # empirical mean and std
        mean = np.sum([w*x for (x, w) in zip(hist_data, hist_weight)])
        std = np.sqrt(np.sum([w*(x**2) for (x, w) in zip(hist_data, hist_weight)]) - mean**2)
        ax.set_title(f"mean: {mean:.3f} ; std: {std:.3f}, \n"
                     f"min: {min(hist_data):.3f} ; max: {max(hist_data):.3f}")
        if xlim is not None:
            ax.set_xlim(*xlim)
            
        return ax
        
    # --- Scenario interpolation --- 
    def nearest_nodes(self, n_nearest, scenario_path, across_tree=True, norm_ord=2, metric=None):
        """Finds the nodes closest to a reference scenario path.
        
        Arguments:
        ----------
        n_nearest: integer >= 1
            The number of nearest neighboors to find.
            
        scenario_path: dictionary mapping stage to scenario
            The stages are integer starting at 0 up to some stage, and the scenario are given as 1darray
            
        across_tree: boolean
            If True, closest nodes are found by comparing the whole scenario path to that of all 
            the nodes at the stage of interest. If False, the comparison is done stage by stage with the scenario
            at the child nodes and moving one stage forward in the tree until the stage of interest is reached and 
            the closest node is found.
        
        norm_ord: float > 0 (default 2)
            The order of the norm considered for the distance (only if `metric` is None).
            
        metric: function mapping a tuple of two 1darray to a positive number
            The metric used to compute the distance (if not provided, then the standard norm is used)
            
        Returns:
        iterator: iterator on the nearest nodes ranked by increasing distance.
        """
        stage = max(scenario_path.keys())
        if stage == 0:
            return [self]
        assert n_nearest <= self.width[stage-1], (f"The number of nearest neighbors ({n_nearest}) is larger than "
                                                    f"the number of nodes at stage {stage}: {self.width[stage-1]}")
        if metric is None:
            metric = lambda x, y: np.linalg.norm(x - y, ord=norm_ord)
            
        if across_tree:
            return self._nn_across_tree(n_nearest, scenario_path, metric)
        else:
            return self._nn_across_children(n_nearest, scenario_path, metric)
     
    def _from_scenario_path_to_numpy(self, scenario_path):
        return np.concatenate([scenario_path[stage][var_name]
                                   for stage in scenario_path.keys()
                                       if scenario_path[stage] is not None
                                           for var_name in scenario_path[stage].keys()])
    
    def _from_scenario_path_to_map_stage_to_rvar_nb(self, scenario_path):
        return {stage: {var_name: len(scenario_path[stage][var_name]) for var_name in scenario_path[stage].keys()}
                            for stage in scenario_path.keys() if scenario_path[stage] is not None}
        
    def _nn_across_tree(self, n_nearest, scenario_path, metric=None):
        """Nearest neighbors across tree"""
        ref_scenario = self._from_scenario_path_to_numpy(scenario_path) # numpy array
        map_stage_to_rvar_nb = self._from_scenario_path_to_map_stage_to_rvar_nb(scenario_path)
        map_stage_to_rvar_names = {stage: list(map_stage_to_rvar_nb[stage].keys()) 
                                            for stage in map_stage_to_rvar_nb.keys()}
        distances = {}
        stage = max(scenario_path.keys())
        for node in self.nodes_at_level(stage):
            node_scenario = self.get_path_as_numpy(node, map_stage_to_rvar_names) # numpy array
            distances[node] = metric(ref_scenario, node_scenario)
        if n_nearest == 1:
            return [min(distances.keys(), key=lambda node: distances[node])]
        else:
            return sorted(distances.keys(), key=lambda node: distances[node])[:n_nearest]
    
    def _nn_across_children(self, n_nearest, scenario_path, metric=None):
        """Nearest neighbors across children"""
        nearest_nodes = []
        scen_tree = self.copy()
        for _ in range(n_nearest):
            node = scen_tree
            while node.level + 1 in scenario_path.keys():
                distances = {child: metric(child.data["scenario"], scenario_path[child.level]) 
                                        for child in node.children}
                node = min(distances, key=lambda child: distances[child])
            nearest_nodes.append(node)
            node.remove_branch()
        return nearest_nodes
                       
    # --- Copy, save, load ---
    def copy(self, deep_copy=False):
        return self.__class__(Node.copy(self, deep_copy))
           
    @classmethod
    def from_file(cls, path, extension):
        return cls(Node.from_file(path, extension))

        
def average(scenario_tree: ScenarioTree, 
            map_stage_to_rvar_names: Optional[Dict[int, List[str]]] = None) -> ScenarioTree:
    """Return a new scenario tree with some scenarios replaced by their average value.
    Note: this function deep copies the input scenario tree so as to not replace the scenarios 'in place' unlike 
    the method `average` of ScenarioTree.
    
    Argument:
    ---------
    map_stage_to_rvar_names: Dict[int, List[str]] or None (default: None)
        The stages (int) and variables names (List[str]) for which the scenarios are averaged.
        
    Returns:
    --------
    instance of ScenarioTree: the scenario tree with some scenarios averaged
    """
    scen_tree = copy.deepcopy(scenario_tree)
    scen_tree.average(map_stage_to_rvar_names)
    return scen_tree  

def decompose(scenario_tree: ScenarioTree) -> List[ScenarioTree]:
    """ Return a list of scenario trees each with a single scenario up to the leaf`"""
    return [combtree_from_scenarios(scenario_tree.get_path_as_numpy(leaf)[np.newaxis], 
                                    scenario_tree.map_stage_to_rvar_nb) for leaf in scenario_tree.leaves] 

def collapse_twostage(tree: ScenarioTree) -> ScenarioTree:
    """Return the scenario tree built from `self` by merging identical scenarios. 
    Note: Works only for two-stage."""
    assert tree.depth == 2, "Scenario tree should be two-stage"
    unique_scenarios, inverse_indices = np.unique(tree.to_numpy(), axis=0, return_inverse=True)
    weights = np.array([child.data["W"] for child in tree.children])
    new_weights = [np.sum(weights[inverse_indices == index]) for index in range(len(unique_scenarios))]
    return twostage_from_scenarios(unique_scenarios, tree.map_stage_to_rvar_nb[1], new_weights)
                
def product(tree1: ScenarioTree, tree2: ScenarioTree):
    """Return the product of two scenario trees (works for two-stage only).
    The product scenario tree represents the joint uncertainty of the two distributions defined 
    by the input scenario trees. Note that only data keys "W" and "scenario" are built in 
    the output tree (all other keys in in the input trees will not be copied)."""
    assert tree1.depth == 2 and tree2.depth == 2, "Scenario trees should be two-stage"
    new_tree = from_bushiness([tree1.width[-1] * tree2.width[-1]])
    new_tree.data["W"] = 1
    for k, child in enumerate(new_tree.children):
        k2, k1 = k // tree1.width[-1], k % tree1.width[-1]
        data1 = tree1.node_at_address((k1,)).data
        data2 = tree2.node_at_address((k2,)).data
        child.data["W"] = data1["W"] * data2["W"]
        child.data["scenario"] = {**data2["scenario"], **data1["scenario"]}
    return new_tree

def _concatenate(tree1, tree2, deep_copy=False):
    final_tree = tree1.copy(deep_copy)
    for leaf in list(final_tree.leaves):
        leaf.add(*tree2.copy(deep_copy).children)
        for node in leaf.nodes:
            if node.address != leaf.address:
                node.data["W"] *= leaf.data["W"]
    return final_tree
            
def concatenate(trees: List[ScenarioTree], deep_copy: Optional[bool] = False):
    assert isinstance(trees, (list, tuple)) and len(trees) >= 2, \
        f"There must be at least 2 scenario trees."
    final_tree  = trees[0].copy(deep_copy)
    for tree in trees[1:]:
        final_tree = _concatenate(final_tree, tree.copy(deep_copy))
    return final_tree
    
from_file = ScenarioTree.from_file
from_topology = ScenarioTree.from_topology
from_recurrence = ScenarioTree.from_recurrence
from_bushiness = ScenarioTree.from_bushiness
from_data_dict = ScenarioTree.from_data_dict
twostage_from_scenarios = ScenarioTree.twostage_from_scenarios
combtree_from_scenarios = ScenarioTree.combtree_from_scenarios
optimal_width_vector = ScenarioTree.optimal_width_vector

                
get_scenario_path = ScenarioTree.get_scenario_path




# -*- coding: utf-8 -*-

from typing import Tuple, Optional, List

import os
import datetime
import itertools
Cartesian = itertools.product
import docplex.mp.model
import numpy as np

from stochoptim.util import RedirectStd
import stochoptim.scenclust.general_partitioning as gp

class CostSpaceScenarioPartitioning:
    """
    Class to perform 'decision-based' scenarios partitioning.
        
    It is an implementation of the Cost Space Scenario Clustering (CSSC) algorithm, which measures the quality 
    of a partition via a score defined in the space of cost values rather than in the space of distributions. 
    It can thus be used as a way to do 'problem-driven' scenario reduction.
    
    For now there are two ways to solve the partitioning problem: one is by solving its mixed-integer programming 
    formulation (typically possible for N smaller than 1000) and the other one is by drawing randomly many partitions
    and picking the best one according to the partitioning score (the lower the better) (possible for any N).
    
    (A heuristic solution method is yet to be developed.)
    """
    
    def __init__(self, opport_cost_matrix: np.ndarray):
        """ 
        Argument:
        ---------
        opport_cost_matrix: 2d-array of shape (N, N) with any N >= 1.
            Square-matrix of opportunity cost values.
            Can be computed via method `compute_opportunity_cost` in module `stochastic_problem_basis`.
        """
        assert len(opport_cost_matrix.shape) == 2 and opport_cost_matrix.shape[0] == opport_cost_matrix.shape[1], \
                "Wrong input dimensions: it should be a square matrix."
        
        self._opport_cost_matrix = opport_cost_matrix
        self._solution_mip = dict()
        self._solution_random = dict()
        
    # --- Properties ---
    @property
    def n_scenarios(self) -> int:
        return len(self._opport_cost_matrix)
    
    @property            
    def solution_mip(self) -> dict:
        return self._solution_mip
        
    @property            
    def solution_random(self) -> dict:
        return self._solution_random
      
    # --- General methods ---
    def best_representatives(self, partition: Tuple[Tuple[int]]) -> Tuple[int]:
        """Returns the best representatives for a given partition (as defined by those minimizing the 
        partition score).
        
        Argument:
        ---------
        partition: tuple containing tuples of integers.
            Integers are in the same tuple if they belong to the same subset. 
            There are as many sub-tuples as there are subsets.
            
        Returns:
        --------
        tuple of integers: the representatives
            Each integer at position k is the representative of the subset at position k in the partition.
        """
        representatives = []
        for subset in partition:
            error = [self.score_subset(subset, p) for p in subset]
            representatives.append(subset[np.argmin(error)])
        return tuple(representatives)
    
    def score_subset(self, subset: Tuple[int], rep: int) -> float:
        """Computes the score of a specific subset in the partition.
        
        Arguments:
        ----------
        subset: tuple of integers
            The elements in the subset.
            
        rep: integer
            The representative of the subset.
            
        Returns:
        --------
        float >= 0: the subset score.
        """
        assert rep in subset, f"The representative {rep} does not belong to the subset {subset}."
        return abs(np.mean(self._opport_cost_matrix[rep, subset]) - self._opport_cost_matrix[rep, rep]) 
    
    def score_partition(self, partition: Tuple[Tuple[int]], representatives: Tuple[int]):
        """Computes the score of the whole partition.
        
        Arguments:
        ----------
        partition: tuple containing tuples of integers
            Integers are in the same tuple if they belong to the same subset. 
            There are as many sub-tuples as there are subsets.
            
        representatives: tuple of integers
            Each integer at position k is the representative of the subset at position k in the partition.
            
        Returns:
        --------
        float >= 0: the score of the whole partition.
        """
        weights = np.array([len(subset) / self.n_scenarios for subset in partition])
        score_partition = [self.score_subset(subset, rep) for subset, rep in zip(partition, representatives)]
        return np.average(score_partition, weights=weights)
   
    # --- MIP exact solver ---
    def solve_mip(self, 
                  cardinality: int, 
                  is_cardinality_fixed: bool = True,
                  timelimit: Optional[int] = None,
                  random_warmstarts: int = 0,
                  warmstart_partition: Tuple[Tuple[int]] = None,
                  logfile: bool = True,
                  **kwargs):
        """Computes the best partition by solving the MIP formulation.
        
        Arguments:
        ----------
        cardinality: int >= 1
            Cardinality of the partition.
             
        is_cardinality_fixed: bool (default: True)
            If True, the cardinality is exactly the one given as argument.
            If False, it is at most the one given as argument.
            
        timelimit: int >= 1 or None (default: None)
            Time limit for the partitioning problem. If None, no limit is set.
            
        random_warmstarts: int >= 0 (default: 0)
            Number of randomly generated partitions from which the best is used as warmstart.
            
        warmstart_partition: tuple containing tuples of integers or None (default: None)
            Partition used as a warmstart. (The representatives of that partition are computed in an optimal way.)

        logfile: bool (default: True)
            If True, the solver's log is saved in a file. This file is located in a folder 'logfiles' in the 
            current directory (the folder is created if it doesn't already exist).
            
        kwargs: 
        -------
        lowerbound: int >= 1 (default: 1)
            The minimum number of elements in each subset of the randomly generated warmstart partitions.
            
        uniform_sampling: bool (default: False)
            If True, the partitions are sampled from the uniform distribution. However, this is much slower. 
            If False, a non-uniform distribution is sampled (this distribution still has a positive probability 
            of selecting any partition).
        """
        if logfile:
            if os.path.exists('logfiles') is False: os.mkdir('logfiles')
            filename = fr'logfiles\solver {str(datetime.datetime.now())}.log'.replace(":", "_")
        else: filename = None

        self._prob = docplex.mp.model.Model()
        self._initialize_decisions()
        self._initialize_constraints(cardinality, is_cardinality_fixed)
        self._prob.minimize(self._objective_function())

        if timelimit is not None:
            self._prob.set_time_limit(timelimit)
         
        if random_warmstarts > 0:
            clustering_score_dict = self._random_partitions_dict(cardinality, random_warmstarts, **kwargs)
            best_partition, best_rep = min(clustering_score_dict.keys(), key=(lambda k: clustering_score_dict[k]))
            dict_warmstart = self._build_warmstart(best_partition, best_rep)
            docplex_warmstart = docplex.mp.solution.SolveSolution(self._prob, dict_warmstart)
            self._prob.add_mip_start(docplex_warmstart)
            
        if warmstart_partition is not None:
            representatives = self.best_representatives(warmstart_partition)
            dict_warmstart = self._build_warmstart(warmstart_partition, representatives)
            docplex_warmstart = docplex.mp.solution.SolveSolution(self._prob, dict_warmstart)
            self._prob.add_mip_start(docplex_warmstart)
            
        with RedirectStd(filename):  
            self._prob.solve(log_output=logfile)
                
        self._build_solution_dict()
        
    def _initialize_decisions(self):
        # t[n] = error of cluster n (possibly 0 if n is not a cluster)
        self._t = self._prob.continuous_var_list(self.n_scenarios, name="t") 
        # x[n1, n2] = 1 if scenario n1 is in cluster n2
        self._x = self._prob.binary_var_matrix(self.n_scenarios, self.n_scenarios, name="x") 
        # u[n] = 1 if scenario n is the representative of a cluster
        self._u = self._prob.binary_var_list(self.n_scenarios, name="u")
        
    def _objective_function(self):
        return self._prob.sum([(1 / self.n_scenarios) * self._t[n] for n in range(self.n_scenarios)])

    def _initialize_constraints(self, cardinality, is_cardinality_fixed=True):
        for n2 in range(self.n_scenarios):
            # error of each cluster (possibly 0 if cluster is 'closed')
            first_sum = self._prob.sum([self._x[n1, n2] * self._opport_cost_matrix[n2, n1] for n1 in range(self.n_scenarios)])
            second_sum = self._prob.sum([self._x[n1, n2] * self._opport_cost_matrix[n2, n2] for n1 in range(self.n_scenarios)])
            self._prob.add_constraint(self._t[n2] >= first_sum - second_sum)
            self._prob.add_constraint(self._t[n2] >= second_sum - first_sum)
        # closed clusters have no scenarios
        self._prob.add_constraints([self._x[n1, n2] <= self._u[n2] for n1 in range(self.n_scenarios)
                                                                    for n2 in range(self.n_scenarios)])
        # a representative must be in the cluster it represents
        self._prob.add_constraints([self._x[n2, n2] == self._u[n2] for n2 in range(self.n_scenarios)])
        # each scenario must be in one cluster
        self._prob.add_constraints([self._prob.sum([self._x[n1, n2] for n2 in range(self.n_scenarios)]) == 1 
                                                        for n1 in range(self.n_scenarios)])
        # limit on the number of clusters
        if is_cardinality_fixed:
            self._prob.add_constraints([self._prob.sum([self._u[n2] 
                                            for n2 in range(self.n_scenarios)]) == cardinality])
        else: 
            self._prob.add_constraints([self._prob.sum([self._u[n2] 
                                            for n2 in range(self.n_scenarios)]) <= cardinality])

    def _build_warmstart(self, partition, representatives):
        """Build a solution dictionary from a partition and its representatives."""
        assert len(representatives) == len(partition), (f"the lengths of the partition {len(partition)} and the "
                                                       f"representatives ({len(representatives)}) do not match.")
        u = {f'u_{n}': 1. if n in representatives else 0 for n in range(self.n_scenarios)}            
        x = {f'x_{n1}_{n2}': 0.  for n1 in range(self.n_scenarios) for n2 in range(self.n_scenarios)}
        for k, n2 in enumerate(representatives):
            for n1 in partition[k]:
                x[f'x_{n1}_{n2}'] = 1.
        first_sum = lambda n2: sum([x[f'x_{n1}_{n2}'] * self._opport_cost_matrix[n2, n1] for n1 in range(self.n_scenarios)])
        second_sum = lambda n2: sum([x[f'x_{n1}_{n2}'] * self._opport_cost_matrix[n2, n2] for n1 in range(self.n_scenarios)])
        t = {f"t_{n2}": abs(first_sum(n2) - second_sum(n2)) for n2 in range(self.n_scenarios)}
        return {**u, **x, **t}

    def _build_solution_dict(self):
        # get partition, representatives, and weights
        representatives = tuple([n for n in range(self.n_scenarios) if self._u[n].solution_value])
        partition = tuple([tuple([n1 for n1 in range(self.n_scenarios) if self._x[n1, n2].solution_value]) 
                            for n2 in representatives])
        score = self._prob.objective_value
        solve_details = self._prob.solve_details
        self._solution_mip = {'partition': partition, 
                              'representatives': representatives, 
                              'weights': self._weights(partition), 
                              'score': score,
                              'time': solve_details.time,
                              'bound': solve_details.best_bound,
                              'status': solve_details.status_code}
         
    def _weights(self, partition):
        return np.array([len(part) / self.n_scenarios for part in partition])
    
    def print_partition(self):        
        print("Clustering score: {:.3g}".format(self._prob.objective_value))
        representatives = tuple([n for n in range(self.n_scenarios) if self._u[n].solution_value])
        partition = tuple([tuple([n1 for n1 in range(self.n_scenarios) if self._x[n1, n2].solution_value]) 
                                            for n2 in representatives])
        for k, rep in enumerate(representatives):
            print(f"{rep} representative of {partition[k]}")
                    
    # --- Random partitioning ---
    def best_random_partition(self, cardinality: int, n_sample: int, **kwargs):
        """ Returns the best partition and representatives out of a number of randomly generated partitions.
        
        Arguments:
        ----------
        cardinality: int
            Cardinality of the partition.
            
        n_sample: int
            The number of partitions randomly generated. 
             
        kwargs: 
        -------
        lowerbound: int >= 1 (default: 1)
            The minimum number of elements in each subset of the partition
            
        uniform_sampling: bool (default: False)
            If True, the partitions are sampled from the uniform distribution. However, this is much slower. 
            If False, a non-uniform distribution is sampled (this distribution still has a positive probability 
            of selecting any partition).
        
        Returns:
        --------
        dictionary: mapping
            'partition': tuple of tuples of ints
            'representatives': tuple of ints
            'score': float
        """
        clustering_score_dict = self._random_partitions_dict(cardinality, n_sample, **kwargs)
        best_partition, best_rep = min(clustering_score_dict.keys(), key=(lambda k: clustering_score_dict[k]))
        best_score = clustering_score_dict[best_partition, best_rep]
        self._solution_random = {'partition': best_partition, 
                                 'representatives': best_rep, 
                                 'weights': self._weights(best_partition),
                                 'score': best_score,
                                 'all scores': np.array(list(clustering_score_dict.values()))}
        
    def _random_partitions_dict(self, cardinality, n_sample, **kwargs):
        """ Returns the best partition and representatives out of a number of randomly generated partitions.
        
        Arguments:
        ----------
        see .get_random_partition()
            
        Returns:
        --------
        dictionary: mapping
            (partition, representatives): clustering score
        """
        clustering_score_dict = {}        
        for index, partition in enumerate(gp.partitioning(list(range(self.n_scenarios)), 
                                                          cardinality, 
                                                          n_sample, 
                                                          **kwargs), 1):
            # pick the cluster representators and their weights  
            representatives = self.best_representatives(partition)
            clustering_score = self.score_partition(partition, representatives)  
            clustering_score_dict[(partition, representatives)] = clustering_score
        return clustering_score_dict    
            
        


# -*- coding: utf-8 -*-

import os
from typing import Dict, Callable, List, Optional, Tuple

import numpy as np
import matplotlib.pyplot as plt
from time import time

try:
    import sklearn.cluster
except ModuleNotFoundError as e:
    print(e)
    
try:
    import pyclustering.cluster.kmedoids
    import pyclustering.cluster.kmedians
except ModuleNotFoundError as e:
    print(e)
    
try:
    import kmodes.kprototypes
except ModuleNotFoundError as e:
    print(e)
    
try:
    import sklearn_extra.cluster
except ModuleNotFoundError as e:
    print(e)  

from stochoptim.scenclust.cost_space_partition import CostSpaceScenarioPartitioning
from stochoptim.scengen.scenario_tree import ScenarioTree
from stochoptim.scengen.decision_process import DecisionProcess
from stochoptim.stochprob.stochastic_problem_basis import StochasticProblemBasis
from stochoptim.stochprob.stochastic_solution_basis import StochasticSolutionBasis
 
class ScenarioClustering:
    """
    Class that wraps together several algorithms to perform scenario reduction on a set of (equal-weight) scenarios 
    for a two-stage stochastic problem (e.g., K-means, K-medoids, Cost-Space-Scenario-Clustering, Monte Carlo, etc.).
    """
    def __init__(self, 
                 scenario_tree: ScenarioTree, 
                 stochastic_problem: StochasticProblemBasis, 
                 is_categorical: Optional[np.ndarray] = None,
                 opport_cost_matrix: Optional[np.ndarray] = None,
                 reference_solution: Optional[StochasticSolutionBasis] = None,
                 load_one_hot_scenarios: bool = False):        
        """        
        Arguments:
        ----------
        scenario_tree: instance of ScenarioTree
            The uncertainty representation of the stochastic problem.

        stochastic_problem: subclass of StochasticProblemBasis
            The two-stage stochastic problem.
            
        is_categorical: 1d-array of shape (n_features,) or None (default: None)
            (where `n_features` is the number of features in one scenario.)
            Array has True at pos i if the i-th feature is a categorical variable; else False.
            
        opportunity_cost_matrix: 2d-array of shape (n_scenarios, n_scenarios) or None (default: None)
            (where `n_scenarios` is the number of scenarios in the scenario tree, i.e., the number of child nodes)
            This matrix is used for the Cost Space Scenario Clustering (CSSC) algorithm. 
            If not provided, the algorithm is not available but the other clustering methods still are.  
            
        reference_solution: instance of StochasticSolutionBasis (or subclass) or None (default: None)
            A solution of the stochastic problem that will served as reference to compute the error gap.
            
        load_one_hot_scenarios: bool (default: False)
            If True and if the scenarios have some categorical features, the scenarios of one-hot representation of these
            features is created. It may then be used by K-means, K-medoids, K-medians, etc. to perform clustering on
            categorical variables. (Note that these scenarios will be much larger than the original because of the one-hot
            representation.)
        """
        # scenarios
        self._scenario_tree = scenario_tree
        self._scenarios = self._scenario_tree.to_numpy()
        self._n_scenarios = len(self._scenarios)
        self._n_features = self._scenarios.shape[1]
        
        # categorical features
        self._is_categorical = is_categorical
        self._unique_cat = set()
        self._n_unique_cat = None
        self._n_num_features = None
        self._n_cat_features = None
        self._scen_num_bin = None
            
        # stochastic problem
        self._stochastic_problem = stochastic_problem
        self._opport_cost_matrix = opport_cost_matrix
        self._reference_solution = reference_solution
        self._map_rvar_to_nb = self._stochastic_problem.map_rvar_name_to_nb[1]
                
        # clustering methods
        self._methods_type_a = ["CSSC", "Kmedoids", "MC"]    # methods with representatives
        self._methods_type_b = ["Kmeans", "Kmedians"]   # methods without representatives
                      
        # Kprototypes only available if array `is_categorical` is provided
        if self._is_categorical is not None:
            self._methods_type_b.append("Kprototypes") 
            
        self._methods_available = self._methods_type_a + self._methods_type_b
        self._results = {}
        self._solutions = {}
        self._start = None
        
        self._set_categorical_attributes(load_one_hot_scenarios)
        self._check_sanity()
        
    def _set_categorical_attributes(self, load_one_hot_scenarios):
        if self._is_categorical is not None:
            # check that categories are all integers
            unique_cat = np.unique(self._scenarios[:, self._is_categorical])
            assert (unique_cat.astype('int') == unique_cat).all(), "Some categories are not integers."
            # set unique categories
            self._unique_cat = set(int(cat) for cat in unique_cat)
            self._n_unique_cat = len(self._unique_cat)
            # numerical features
            scen_num = self._scenarios[:, ~self._is_categorical]
            self._n_num_features = scen_num.shape[1]
            # categorical categorical 
            scen_cat = self._scenarios[:, self._is_categorical]
            self._n_cat_features = scen_cat.shape[1]
            if load_one_hot_scenarios:
                # categorical to binary with more features (one-hot encoding)
                scen_bin_3d = {cat: np.ones((self._n_scenarios, self._n_cat_features)) for cat in self._unique_cat}
                scen_bin_3d = {cat: (scen_bin_3d[cat] == scen_cat).astype('int') for cat in self._unique_cat}
                scen_bin = np.concatenate([scen_bin_3d[cat] for cat in self._unique_cat], axis=1)
                self._scen_num_bin = np.concatenate([scen_num, scen_bin], axis=1)
        else:
            self._n_unique_cat = 0
            self._n_num_features = self._n_features
            self._n_cat_features = 0
            
    # --- Properties ---
    @property
    def results(self):
        return self._results
    
    @property
    def solutions(self):
        return self._solutions
    
    @property
    def scenarios(self):
        return self._scenarios
    
    @property
    def n_features(self):
        """Total number of features in a scenario (i.e., number of parameters included in a scenario)."""
        return self._n_features

    @property
    def n_runs(self) -> Dict[str, Dict[int, int]]:
        return {method: {card: self.get_n_runs(method, card) for card in self.cardinalities(method)}
                                 for method in self._methods_available}
        
    def cardinalities(self, method):
        return sorted([card for meth, card in self._results.keys() if meth == method])
    
    def n_random_features(self, method=None, cardinality=None, index_sample=0):
        """Number of actually random features in the original set of scenarios. A feature is random if it 
        has a standard deviation > 0."""
        n_rand_features = lambda scenarios: len((np.max(scenarios, axis=0) - np.min(scenarios, axis=0)).nonzero()[0])
        if method is None:
            return n_rand_features(self._scenarios)
        else:
            return n_rand_features(self.get_scenarios(method, cardinality, index_sample))

    # --- Clustering methods ---       
    def run_cssc(self, 
                 cardinality, 
                 is_cardinality_fixed=True, 
                 solve_mip=True, 
                 n_random_warmstarts=10**3, 
                 timelimit=None, 
                 **kwargs):
        """Scenario clustering by the Cost Space Scenario Clustering (CSSC) method (implemented in the module 
        `cost_space_partition`).
        
        Arguments:
        ----------
        cardinality: int >= 1 
            The number of clusters in the partition.
            
        solve_mip: bool (default: True)
            If True, the CSSC problem is tackled via the MIP formulation solved using an exact solver.
            If False, the CSSC problem is tackled by sampling partitions randomly and picking the best one (in that 
            case the `warmstart` argument must be >= 1).
            
        n_random_warmstarts: int >= 0 (default: 10**3)
            Number of randomly generated partitions out of which the one with the lowest clustering score will 
            be picked as the best partition or used to initialize the MIP formulation. 
            
        timelimit: int >= 1 or None (default: None)
            Time limit for the partioning problem. If None, no limit is set.
            
        kwargs:
        -------
        logfile: bool (default: True)
            If True, the solver's log is saved in a file. This file is located in a folder 'logfiles' in the 
            current directory (the folder is created if it doesn't already exist).
                              
        warmstart_partition: tuple of tuples of int, or None (default: None)
            Must be None if warmstart is not None.
            If provided, this partition is used as warmstart (along with the representatives).
            
        lowerbound: int >= 1 (default: 1)
            The minimum number of elements in each subset of the randomly generated warmstart partitions.
            
        uniform_sampling: bool (default: False)
            If True, the partitions are sampled from the uniform distribution. However, this is much slower. 
            If False, a non-uniform distribution is sampled (this distribution still has a positive probability 
            of selecting any partition).
        """
        assert self._opport_cost_matrix is not None, "Provide the opportunity cost matrix to run the CSSC algorithm."
        self._check_opportunity_cost_matrix()
        assert solve_mip or n_random_warmstarts >= 1, "`n_random_warmstarts` should be >= 1 if solve_mip=False"
        self.cssc = CostSpaceScenarioPartitioning(self._opport_cost_matrix)
        self._start = time()
        
        if solve_mip:
            self.cssc.solve_mip(cardinality, is_cardinality_fixed, timelimit, n_random_warmstarts, **kwargs)
            representatives = self.cssc.solution_mip['representatives']
            partition = self.cssc.solution_mip['partition']
            score = self.cssc.solution_mip['score']
            
        else:
            self.cssc.best_random_partition(cardinality, n_random_warmstarts, **kwargs)
            representatives = self.cssc.solution_random['representatives']
            partition = self.cssc.solution_random['partition']
            score = self.cssc.solution_random['score']
        self._set_results('CSSC', len(representatives), representatives, partition, None, score)
    
    def run_kmeans(self, cardinality, **kwargs):
        """Scenario clustering by Kmeans method."""
        self._start = time()
        self.kmeans = sklearn.cluster.KMeans(n_clusters=cardinality, **kwargs)
        
        if self._n_unique_cat > 0: # if some categorical features, clustering done on one-hot encoding of features
            self.kmeans.fit(self._scen_num_bin)
            num_centroids = self.kmeans.cluster_centers_[:, :self._n_num_features]
            bin_centroids = self.kmeans.cluster_centers_[:, -(self._n_cat_features * self._n_unique_cat):]
            cat_centroids = self._one_hot_centroids_to_cat(bin_centroids)
            self.kmeans.scenarios = np.zeros((cardinality, self.n_features))
            self.kmeans.scenarios[:, ~self._is_categorical] = num_centroids
            self.kmeans.scenarios[:, self._is_categorical] = cat_centroids
        else:
            self.kmeans.fit(self._scenarios)
            self.kmeans.scenarios = np.array(self.kmeans.cluster_centers_)
        self.kmeans.partition = tuple(tuple((self.kmeans.labels_ == i).nonzero()[0]) for i in range(cardinality))
        self._set_results('Kmeans', cardinality, None, self.kmeans.partition, 
                          self.kmeans.scenarios, self.kmeans.inertia_)
        
    def run_kmedians(self, cardinality, **kwargs):
        """Scenario clustering by the Kmedian algorithm of 'pyclustering'"""
        self._start = time()
        random_indices = np.random.choice(range(self._n_scenarios), size=cardinality, replace=False)
        
        if self._n_unique_cat > 0: # if some categorical features, clustering done on one-hot encoding of features
            initial_medians = self._scen_num_bin[random_indices]
            self.kmedians = pyclustering.cluster.kmedians.kmedians(self._scen_num_bin, initial_medians, **kwargs)
            self.kmedians.process()
            num_centroids = np.array(self.kmedians.get_medians())[:, :self._n_num_features]
            bin_centroids = np.array(self.kmedians.get_medians())[:, -(self._n_cat_features * self._n_unique_cat):]
            cat_centroids = self._one_hot_centroids_to_cat(bin_centroids)
            self.kmedians.scenarios = np.zeros((cardinality, self.n_features))
            self.kmedians.scenarios[:, ~self._is_categorical] = num_centroids
            self.kmedians.scenarios[:, self._is_categorical] = cat_centroids
        else:
            initial_medians = self._scenarios[random_indices]
            self.kmedians = pyclustering.cluster.kmedians.kmedians(self._scenarios, initial_medians, **kwargs)
            self.kmedians.process()
            self.kmedians.scenarios = np.array(self.kmedians.get_medians())

        self.kmedians.partition = tuple(tuple(part) for part in self.kmedians.get_clusters())
        self.kmedians.score = self.kmedians.get_total_wce()
        self._set_results('Kmedians', cardinality, None, self.kmedians.partition, 
                          self.kmedians.scenarios, self.kmedians.score)

    def run_kmedoids(self, cardinality, which_kmedoids='pyclustering', **kwargs): 
        """Scenario clustering by the Kmedoids algorithm of the 'pyclustering' or 'sklearn_extra' library.
        
        Arguments:
        ----------
        which_kmedoids: {'pyclustering', 'sklearn_extra'}
        """
        param = {**{'which_kmedoids': which_kmedoids}, **kwargs}
        self._start = time()
        if which_kmedoids == 'pyclustering':
            initial_medoids = list(np.random.choice(range(self._n_scenarios), size=cardinality, replace=False))
            if self._n_unique_cat > 0: # if some categorical features, clustering done on one-hot encoding of features
                self.kmedoids = pyclustering.cluster.kmedoids.kmedoids(self._scen_num_bin, initial_medoids, **kwargs)
            else:
                self.kmedoids = pyclustering.cluster.kmedoids.kmedoids(self._scenarios, initial_medoids, **kwargs)
            self.kmedoids.process()
            self.kmedoids.score = None
            self.kmedoids.partition = tuple([tuple(part) for part in self.kmedoids.get_clusters()])
            self.kmedoids.representatives = tuple(self.kmedoids.get_medoids()) 
            
        elif which_kmedoids == 'sklearn_extra':
            self.kmedoids = sklearn_extra.cluster.KMedoids(n_clusters=cardinality, **kwargs)
            if self._n_unique_cat > 0: # if some categorical features, clustering done on one-hot encoding of features
                self.kmedoids.fit(self._scen_num_bin)
            else:
                self.kmedoids.fit(self._scenarios)
            self.kmedoids.score = self.kmedoids.inertia_
            self.kmedoids.representatives = tuple(self.kmedoids.medoid_indices_)
            self.kmedoids.partition = tuple(tuple((self.kmedoids.labels_ == i).nonzero()[0]) for i in range(cardinality))
        else:
             raise ValueError("Wrong `which_kmedoids` argument: should be 'pyclustering' or 'sklearn_extra', "
                              f"not {which_kmedoids}")
        self._set_results('Kmedoids', cardinality, self.kmedoids.representatives, self.kmedoids.partition,
                          score=self.kmedoids.score, param=param)
        
    def run_kprototypes(self, cardinality, init='Huang', **kwargs):
        """Scenario clustering by Kprototypes method.
        
        init: {'Huang', 'Cao', 'random'} (default: 'Huang')
            If 'random', both the centroids of the numerical and categorical variables are initialized by picking a 
            scenario randomly.
            If not 'random', the categorical centroids are initialized by 'Huang' or 'Cao' and the numerical by
            k-means++.
        """
        assert self._is_categorical is not None, \
            "Provide the mask on the categorical variables using `is_categorical`" 
        self._start = time()
        
        if self._n_cat_features < self.n_features: # use Kprototype
            self.kprot = kmodes.kprototypes.KPrototypes(n_clusters=cardinality, init=init, **kwargs)
            categorical_indices = self._is_categorical.nonzero()[0]
            self.kprot.fit(self._scenarios, categorical=list(categorical_indices))
            self.kprot.partition = tuple(tuple((self.kprot.labels_ == i).nonzero()[0]) for i in range(cardinality))
            self.kprot.scenarios = -np.ones((cardinality, self._scenarios.shape[1]))
            self.kprot.scenarios[:, ~self._is_categorical] = self.kprot.cluster_centroids_[0]
            self.kprot.scenarios[:, self._is_categorical] = self.kprot.cluster_centroids_[1]
            
        else: # use Kmodes instead of Kprototype
            self.kprot = kmodes.kmodes.KModes(n_clusters=cardinality, init=init, **kwargs)
            self.kprot.fit(self._scenarios)
            self.kprot.scenarios = np.array(self.kprot.cluster_centroids_)
            self.kprot.partition = tuple(tuple((self.kprot.labels_ == i).nonzero()[0]) for i in range(cardinality))
            
        self._set_results('Kprototypes', cardinality, None, self.kprot.partition, 
                          self.kprot.scenarios, self.kprot.cost_)

    def run_monte_carlo(self, cardinality):   
        """Random scenario clustering by Monte Carlo method."""
        self._start = time()
        self.mc_representatives = np.random.choice(range(len(self._scenarios)), size=cardinality, replace=False)
        self._set_results('MC', cardinality, tuple(self.mc_representatives))
        
    # --- Clustering results ---
    def get_n_runs(self, method, cardinality):
        """Return the number of clustering samples performed"""
        return len(self._results.get((method, cardinality), []))
    
    def get_scenarios(self, method, cardinality, index_sample=0):
        try:
            if method in self._methods_type_a:
                return np.array([self._scenarios[rep] for rep in self.get_representatives(method, cardinality, 
                                                                                             index_sample)])
            elif method in self._methods_type_b:
                return self._results[method, cardinality][index_sample]['scenarios']
        except (KeyError, IndexError):
            return None
        
    def get_weights(self, method, cardinality, index_sample=0):
        try:
            return self._results[method, cardinality][index_sample]['weights']
        except (KeyError, IndexError):
            return None
    
    def get_representatives(self, method, cardinality, index_sample=0):
        try:
            return self._results[method, cardinality][index_sample]['reps']
        except (KeyError, IndexError):
            return None

    def get_partition(self, method, cardinality, index_sample=0):
        try:
            return self._results[method, cardinality][index_sample]['partition']
        except (KeyError, IndexError):
            return None
        
    def get_clust_time(self, method, cardinality, index_sample=0):
        try:
            return self._results[method, cardinality][index_sample]['time']
        except (KeyError, IndexError):
            return None
        
    def get_score(self, method, cardinality, index_sample=0):
        try:
            return self._results[method, cardinality][index_sample]['score']
        except (KeyError, IndexError):
            return None
        
    def get_param(self, method, cardinality, index_sample=0):
        try:
            return self._results[method, cardinality][index_sample]['param']
        except (KeyError, IndexError):
            return dict()
        
    def _set_results(self, method, cardinality, representatives, partition=None, scenarios=None, score=None, param=None):
        if self.get_n_runs(method, cardinality) == 0:
            self._results[method, cardinality] = []
        # check param
        assert isinstance(param, dict) or param is None, f"`param` must be a dictionary or None, not {param}"
        param = param if param is not None else dict()
        if method in self._methods_type_a:
            self._set_results_type_a(method, cardinality, representatives, partition, score, param)
        elif method in self._methods_type_b:
            self._set_results_type_b(method, cardinality, partition, scenarios, score, param)
        else:
            raise ValueError(f"{method} is neither of type 'a' nor of type 'b'.")
            
    def _set_results_type_a(self, method, cardinality, representatives, partition, score, param):
        """Set results for methods of type 'a', i.e., those with representatives (e.g., CSSC, Kmedoids, MC)."""
        if partition is None: # equal weights
            self._results[method, cardinality].append({'reps': representatives,
                                                      'weights': np.ones((cardinality,)) / cardinality,
                                                      'partition': None,
                                                      'score': score,
                                                      'param': param,
                                                      'time': time() - self._start})
        else:
            self._results[method, cardinality].append({'reps': representatives,
                                                      'weights': np.array([len(part) / self._n_scenarios 
                                                                               for part in partition]),
                                                      'partition': partition,
                                                      'score': score,
                                                      'param': param,
                                                      'time': time() - self._start})
        
    def _set_results_type_b(self, method, cardinality, partition, scenarios, score, param):
        """Set results for methods of type 'b', i.e., those with partition but no representatives (e.g., Kmeans)"""
        self._results[method, cardinality].append({'scenarios': scenarios,
                                                  'weights': np.array([len(part) / self._n_scenarios 
                                                                           for part in partition]),
                                                  'partition': partition,
                                                  'score': score,
                                                  'param': param,
                                                  'time': time() - self._start})
        
    def _one_hot_centroids_to_cat(self, binary_centroids: np.ndarray):
        """
        Argument:
        ---------
        binary_centroids: 2d-array of shape (cardinality, n_cat * n_cat_features)
            Centroids of the one-hot encoding of categories (with continuous values inside the interval [0, 1])
         
        Returns:
        --------
        cat_centroids: 2d-array of shape (cardinality, n_cat_feeatures)
            Centroids of the categories (with values in `unique_cat`)
        """
        cardinality = binary_centroids.shape[0]
        assert binary_centroids.shape[1] == self._n_unique_cat * self._n_cat_features, \
            (f"Wrong shape for binary centroids, should be ({cardinality}, {self._n_unique_cat * self._n_cat_features}), not "
             f"{binary_centroids.shape}")
        bin_centroids_3d = np.zeros((self._n_unique_cat, cardinality, self._n_cat_features))
        for k, cat in enumerate(self._unique_cat):
            bin_centroids_3d[k] = binary_centroids[:, :self._n_cat_features]
            binary_centroids = binary_centroids[:, self._n_cat_features:]
        cat_centroids = bin_centroids_3d.argmax(axis=0) 
        # turn category index (k) to actual category (cat)
        for k, cat in enumerate(self._unique_cat):
            cat_centroids[(cat_centroids == k)] = cat
        return cat_centroids
    
    def delete_result(self, method, cardinality, indices_sample):
        """Delete all results that are associated to a specific clustering method, cardinality and sample index.
        
        This deletes the results of the clustering instance as well as all the stochastic problem solutions 
        obtained for that instance.

        Arguments:
        ----------
        indices_sample: int or list of ints
            The sample index (or indices) to be deleted.
        """
        if isinstance(indices_sample, int):
            indices_sample = [indices_sample]
        assert len(set(indices_sample)) == len(indices_sample), "Each index should appear only once."
        for index in reversed(sorted(indices_sample)): # remove from larger to lower
            # delete from results attribute
            self._results[method, cardinality].pop(index)
            # delete from solution attribute
            self._solutions.pop((method, cardinality, index), None)
           
    def delete_solution(self, method, cardinality, index_sample, timelimit):
        """Delete the solution associated to a specific method, cardinality, sample index and time limit.
        
        Arguments:
        ----------
        timelimit: int/float or 2-tuple of ints/floats
        """
        # delete solution with this timelimit (whether it is a clustered or implementation solution)
        self._solutions.get((method, cardinality, index_sample), {}).pop(timelimit, None)
        # if clustered solution, then delete also the implementation solution obtained from it
        if isinstance(timelimit, (float, int)):
            for tlimit in list(self._solutions.get((method, cardinality, index_sample), {}).keys()):
                if isinstance(tlimit, tuple) and tlimit[0] == timelimit:
                    self._solutions[method, cardinality, index_sample].pop(tlimit, None) 
            
    # --- Solve stochastic problems ---
    def inner_solve(self, 
                      cardinality: int,
                      methods: Optional[List[str]] = None, 
                      indices_sample: Optional[Dict[str, List[int]]] = None, 
                      timelimit_opt: Optional[int] = None, 
                      timelimit_eval: Optional[int] = None, 
                      logfile_opt: Optional[str] = None,
                      logfile_eval: Optional[str] = None,
                      **kwargs):
        """Solve the stochastic problem over the clustered set of scenarios and evaluate the output decisions. 
        
        All solutions are done via the `.solve` method of the problem; if this method does not fit the needs, check 
        the `outer_solve` where one can implement their own 'optimizer' and 'evaluator' function. 
        
        Arguments:
        ----------
        cardinality: int
            The cardinality of the clustering for which the stochatic problem will be solved.
            
        methods: list of str or None (default: None)
            The clustering methods for which the stochastic problem will be solved.
            If None, all methods are solved.
        
        indices_sample: Dict[str, List[int]] or None (default: None)
            Mapping from the method to the indices of samples for which the stochastic problem will be solved.
            If None, all samples are solved.
            
        timelimit_opt: int >= 1 or None (default: None)
            Time limit to solve the proxy problem with the clustered scenarios.
            
        timelimit_eval: int >= 1 or None (default: None)
            Time limit to solve the problem that evaluates the solution of the clustered problem.
            
        logfile_opt: str or None (default: "")
            This string is appended to the log filename of the optimizer. This file is located in a folder 'logfiles' 
            in the current directory (the folder is created if it doesn't already exist).
            If None, no log is available. 
            
        logfile_eval: str or None (default: "")
            This string is appended to the log filename of the evaluator. This file is located in a folder 'logfiles' 
            in the current directory (the folder is created if it doesn't already exist). 
            If None, no log is available. 
            
        kwargs:
        -------
        all kwargs of StochasticProblemBasis.solve() except:
            `timelimit`, `logfile`, `clear_between_trees`
        """
        methods = self._check_and_set_methods(methods)
        indices_sample = self._check_and_set_samples(indices_sample, methods, cardinality)
                   
        method_samples = [(method, index) for method in methods for index in indices_sample[method]]
        # create list of reduced scenarios and weights using all the methods 
        scenarios_sets = [self.get_scenarios(method, cardinality, index) for method, index in method_samples]
        weights_sets = [self.get_weights(method, cardinality, index) for method, index in method_samples]
        scenario_trees = [ScenarioTree.twostage_from_scenarios(scenarios, self._map_rvar_to_nb, weights)
                                                    for scenarios, weights in zip(scenarios_sets, weights_sets)]

        # solve the stochastic problem on the reduced sets
        solutions = self._stochastic_problem.solve(*scenario_trees,
                                                  timelimit=timelimit_opt,
                                                  clear_between_trees=True,
                                                  logfile=logfile_opt,
                                                  fill_scenario_tree=True,
                                                  **kwargs)
        if len(scenario_trees) == 1:
            solutions = [solutions]
        
        for i, (method, index) in enumerate(method_samples):
            key = (method, cardinality, index)
            if self._solutions.get(key) is None:
                self._solutions[key] = {}
            self._solutions[key][timelimit_opt] = solutions[i]
            
        # get the implementation solution
        for method, index in method_samples:
            key = (method, cardinality, index)
            decision_process = DecisionProcess(self._stochastic_problem.map_dvar_to_index, 
                                               {0: self._solutions[key][timelimit_opt].x0})            
            self._solutions[key][(timelimit_opt, timelimit_eval)] = \
                    self._stochastic_problem.solve(self._scenario_tree,
                                                  decision_process=decision_process,
                                                  timelimit=timelimit_eval,
                                                  logfile=logfile_eval,
                                                  **kwargs)
            
    def outer_solve(self,
                     cardinality: int,
                     optimize_fct: Callable[[ScenarioTree], Tuple[StochasticSolutionBasis, DecisionProcess, int]],
                     evaluate_fct: Callable[[ScenarioTree, DecisionProcess], Tuple[StochasticSolutionBasis, int]],
                     methods: Optional[List[str]] = None, 
                     indices_sample: Optional[Dict[str, List[int]]] = None,
                     **kwargs):
        """Solve the stochastic problem over the clustered set of scenarios and evaluate the solution using the optimizer 
        and evaluator functions provided as input.
         
        Arguments:
        ----------
        cardinality: int
            The cardinality of the clustering for which the stochatic problem will be solved.
            
        optimize_fct: The optimizer function that takes a scenario tree (ScenarioTree) and 
            returns a solution to the stochastic problem (StochasticSolutionBasis), the decision process 
            (DecisionProcess) to be evaluated in the next step, and the time (int) it took to solve the problem.
            
        evaluate_fct: The evaluator function that takes a scenario tree (ScenarioTree), a decision process 
            (DecisionProcess), and returns a solution to the problem (StochasticSolutionBasis) with 
            the time (int) it took to solve the problem.
            
        methods: list of str or None (default: None)
            The clustering methods for which the stochastic problem will be solved.
            If None, all methods are solved.
        
        indices_sample: Dict[str, List[int]] or None (default: None)
            Mapping from the method to the indices of samples for which the stochastic problem will be solved.
            If None, all samples are solved.
        """
        methods = self._check_and_set_methods(methods)
        indices_sample = self._check_and_set_samples(indices_sample, methods, cardinality)

        for method in methods:            
            for index_sample in indices_sample[method]:
               
                key = (method, cardinality, index_sample)
                if self._solutions.get(key) is None:
                    self._solutions[key] = {}     
                    
                scenarios = self.get_scenarios(*key) 
                weights = self.get_weights(*key)
                scenario_tree = ScenarioTree.twostage_from_scenarios(scenarios, self._map_rvar_to_nb, weights)
                
                sol_opt, dec_pro, time_opt = optimize_fct(scenario_tree)
                sol_eval, time_eval = evaluate_fct(self._scenario_tree, dec_pro)
                
                self._solutions[key][time_opt] = sol_opt
                self._solutions[key][(time_opt, time_eval)] = sol_eval
        
    def _check_and_set_samples(self, indices_sample, methods, cardinality):
        """Check and set indices_sample input"""
        if indices_sample is None:
            indices_sample = {method: range(self.get_n_runs(method, cardinality)) for method in methods}
        else:
            # check methods name
            assert set(indices_sample.keys()).issubset(set(self._methods_available)), \
                f"Methods {set(methods).difference(set(self._methods_available))} in `indices_sample` do not exist."
            # check indices sample
            for method in methods:
                all_samples = range(self.get_n_runs(method, cardinality))
                if indices_sample.get(method) is not None:
                    assert set(indices_sample[method]).issubset(set(all_samples)), \
                        (f"Sample indices {set(indices_sample[method]).difference(set(all_samples))} do not exist "
                         f"for {method} and cardinality {cardinality}.")
                else:
                   indices_sample[method] = all_samples
        return indices_sample
   
    def _check_and_set_methods(self, methods):
        """Check and set methods input"""
        if methods is None:
            methods = self._methods_available
        else:
            assert isinstance(methods, list), f"`methods` must be of type list, not {type(methods)}."
            # check methods name
            assert set(methods).issubset(set(self._methods_available)), \
                f"Methods {set(methods).difference(set(self._methods_available))} do not exist."
        return methods
                
    def get_opt_solution(self, method, cardinality, index_sample=None, timelimit=None):
        try:
            if index_sample is None:
                index_sample = 0
            solutions = self._solutions[method, cardinality, index_sample]
            if timelimit is None:
                timelimit = [t for t in solutions.keys() if isinstance(t, (int, float)) or t is None][0]
            return solutions[timelimit]
        except (KeyError, IndexError):
            return None
        
    def get_eval_solution(self, method, cardinality, index_sample=None, timelimit_tuple=None):
        try:
            if index_sample is None:
                index_sample = 0
            solutions = self._solutions[method, cardinality, index_sample]
            if timelimit_tuple is None:
                timelimit_tuple = [t for t in solutions.keys() if isinstance(t, tuple)][0]
            return solutions[timelimit_tuple]
        except (KeyError, IndexError):
            return None

    def get_gap(self, method, cardinality, index_sample=None, timelimit_tuple=None):
        """Return the gap (in %) from the reference solution"""
        try:
            v_ref = self._reference_solution.objective_value
            v_sol = self.get_eval_solution(method, cardinality, index_sample, timelimit_tuple).objective_value
            gap = round(100 * (v_ref - v_sol) / (10**-10 + abs(v_ref)), 5)
            return gap if self._stochastic_problem.objective_sense == 'max' else -gap
        except (AttributeError, KeyError):
            return None

    # --- Display Results ---
    def plot_method(self, 
                    method: str,
                    to_show: str = 'obj',
                    samples: Optional[Dict[int, List[int]]] = None, 
                    card_fct: Callable[[int], bool] = lambda card: True,
                    x_in_percent: bool = False,
                    show_mean=False,
                    title: Optional[str] = None,
                    figsize=(7,5),
                    ax=None):
        """
        samples: mapping between the cardinality and the list of sample indices to be plotted.
        
        to_show: {'gap', 'obj', 'clust-time', 'solve-time'}
        """        
        if ax is None:
            fig, ax = plt.subplots(figsize=figsize)
        
        if samples is None:
            samples = {card: range(n_sample) for card, n_sample in self.n_runs[method].items()}
        if show_mean: 
            y_list, card_list = [], []
        for card in sorted(list(samples.keys())):
            if not card_fct(card):
                continue 
            if show_mean: 
                y_list.append([])
                card_list.append(card)
            for index in samples[card]:
                key = (method, card, index)              
                if self._solutions.get(key) is None:
                    continue
                timelimits = self._solutions[key].keys()
                timelimits_tuple = [time for time in timelimits if isinstance(time, tuple)]
                for timelimit in timelimits_tuple:
                    # pick what to display
                    if to_show == 'gap':
                        assert self._reference_solution is not None, \
                            "Impossible to plot the gap if the reference solution is not provided"
                        y = self.get_gap(method, card, index, timelimit)
                        ax.axhline(0)
                    elif to_show == 'obj':
                        y = self._solutions[key][timelimit].objective_value
                        if self._reference_solution is not None:
                            ax.axhline(self._reference_solution.objective_value)
                    elif to_show == 'clust-time':
                        y = self.get_clust_time(*key)
                    elif to_show == 'solve-time':
                        y = self._solutions[key][timelimit[0]].scenario_tree.data['time']
                        if self._reference_solution is not None:
                            ax.axhline(self._reference_solution.scenario_tree.data['time'])
                    else:
                        raise NotImplementedError("Wrong `to_show` argument: should be 'gap', 'obj', 'clust-time', "
                                                  f"or 'solve-time', not {to_show}")
                    if x_in_percent:
                        ax.scatter(100 * card / self._n_scenarios, y, marker="x")
                    else:
                        ax.scatter(card, y, marker="x")
                    if show_mean:
                        y_list[-1].append(y)
        if show_mean:
            ax.plot(card_list, [np.mean(ys) for ys in y_list])
        if x_in_percent:
            ax.set_xlabel("cardinality (%)")
        else:
            ax.set_xlabel("cardinality")
        ax.set_ylabel(to_show)
        if title:
            ax.set_title(title)
        else:
            ax.set_title(f"{method}")
        return ax

    def plot_results(self, methods=None, sharey=True, sharex=False, figsize=(15,3), **kwargs):
        methods = self._methods_available if methods is None else methods
        methods_with_data = [method for method in methods if self.n_runs[method] != dict()]
        fig, ax = plt.subplots(1, len(methods_with_data), figsize=figsize, sharey=sharey, sharex=sharex)
        if len(methods_with_data) == 1:
            self.plot_method(methods_with_data[0], ax=ax, **kwargs)
        else:
            for i, method in enumerate(methods_with_data):
                self.plot_method(method, ax=ax[i], **kwargs)
        return ax
        
    # --- Save and load results ----
    def to_file(self, wdir, tree_extension='pickle', show_files=True, remove_tree_keys=None):
        """Save the clustering instance (clustering results and clustered solution of the stochastic problem).

        Argument:
        ---------
        wdir: string
        
        tree_extension: {'txt', 'pickle'} (default: 'pickle')
        
        show_file: bool (default: True)
        
        remove_tree_keys: List[str] or None (default: None)
            Data keys that will not be saved in the scenario tree in addition to those that are not saved by default
            (see `_save_tree()`).
        """
        # Create folder if doesn't exist
        if not os.path.exists(wdir): 
            os.mkdir(wdir)
        else:
            print(f"Directory already exists. Files may be overwritten.")
        if show_files:
            print(f"These files have been saved at {wdir}:")
        
        # (1) Save string representation just for readable information
        filename = "clustering summary.txt"
        with open(wdir + filename, "w") as f:
            f.write(self.__repr__())
        if show_files:
            print(f"  {filename}")    
        
         # (2) Save result dictionary
        import pickle
        filename = "results.pickle"
        with open(wdir + filename, "wb") as f:
            pickle.dump(self._results, f)
        if show_files:
            print(f"  {filename}")
        
        # (3) Save scenario trees
        for method, cardinality, index_sample in self._solutions.keys():
            key = (method, cardinality, index_sample)
            for timelimit in self._solutions[key].keys():
                filename = self._save_tree(wdir, tree_extension, method, cardinality, index_sample, 
                                           timelimit, remove_tree_keys)
                if show_files:
                    print(f"  {filename}")

    def _save_tree(self, wdir, tree_extension, method, cardinality, index_sample, timelimit, remove_tree_keys):
        if remove_tree_keys is None:
            remove_tree_keys = []
        key = (method, cardinality, index_sample)
        if isinstance(timelimit, (int, float)) or timelimit is None: # tree from clustered problem
            scenario_tree = self._solutions[key][timelimit].scenario_tree
            filename = f"tree_clust-{method}-K{cardinality}-#{index_sample}-{timelimit}sec"
            scenario_tree.to_file(wdir + filename, 'txt', 
                                  without_keys=['scenario', 'memory', 'decision', 'W', 'w'] + remove_tree_keys)
        elif isinstance(timelimit, tuple): # tree from implementation problem
            scenario_tree = self._solutions[key][timelimit].scenario_tree
            filename = f"tree_impl-{method}-K{cardinality}-#{index_sample}-{timelimit}sec"
            scenario_tree.to_file(wdir + filename, tree_extension, 
                                  without_keys=['scenario', 'memory'] + remove_tree_keys)
        else:
            raise TypeError(f"Wrong type for for `timelimit`. Should be int, float or tuple, not {type(timelimit)}")
        return filename
    
    def from_file(self, wdir, tree_extension='pickle', show_files=False):
        """Load the clustering results.
        
        Specifically, this builds the `results` and `solution` attributes of the instance. This is done by
        loading the 'results.pickle' file and all the scenario-tree files saved in the working directory.
        
        Arguments:
        ---------
        wdir: string
            Path to the working directory that contains all the files.
            
        tree_extension: {'txt', 'pickle'} (default: 'pickle')
            Format in which the scenario trees of the implementation problems are saved. 
            (Note that the format for the scenario trees of the clustered problems are always 'txt'.)
        """
        from os import listdir
        from os.path import isfile, join
        import pickle
        
        # (1) Load result dictionary
        loaded_files = ["results.pickle"]  
        try:
            with open(wdir + "results.pickle", "rb") as f:
                self._results = pickle.load(f)
            if show_files:
                print(f"These files have been loaded from {wdir}: \n  results.pickle\n")
                
        except FileNotFoundError:
            print(f"File '{wdir}results.pickle' not found. Impossible to load the clustering results.")
            return 
        
        # (2) Load scenario-tree solution
        files_in_wdir = [f for f in listdir(wdir) if isfile(join(wdir, f))] # get all file in working directory
        for file in files_in_wdir:
            # if file doesn't contain a clustered or implementation scenario tree
            if 'tree_clust' not in file and 'tree_impl' not in file: 
                continue
            loaded_files.append(file)
            # load tree
            filename = file.split('.')[0] # remove extension
            extension = 'txt' if 'clust' in filename else tree_extension
            scenario_tree = ScenarioTree.from_file(wdir + filename, extension)
            
            # get tuple (method, cardinality, index_sample, timelimit)
            method, cardinality, index_sample, timelimit = ScenarioClustering._parse_filename(filename)
            key = (method, cardinality, index_sample)
            if self._solutions.get(key) is None:
                self._solutions[key] = {}
                
            if 'clust' in filename: 
                self._solutions[key][timelimit] = self._stochastic_problem._solution_class(self._stochastic_problem, 
                                                                                        scenario_tree)
            else:
                 # append scenarios to the scenario tree
                scenario_tree.append_data_dict(self._scenario_tree.get_data_dict(['scenario']))
                self._solutions[key][timelimit] = self._stochastic_problem._solution_class(self._stochastic_problem, 
                                                                                       scenario_tree)
            if show_files:
                print(file)
    
    @staticmethod
    def _parse_filename(filename):
        method, cardinality_str, index_sample_str, timelimit_str = filename.split('-')[1:]
        cardinality = int(cardinality_str[1:]) # remove prefix 'K' and convert to integer
        index_sample = int(index_sample_str[1:]) # remove prefix '#' and convert to integer
        timelimit = eval(timelimit_str[:-3]) # remove suffix 'sec' and convert to tuple or integer via eval()
        return method, cardinality, index_sample, timelimit
      
    # --- Representations ---    
    def __str__(self):
        string = ("Scenario Clustering: \n"
                  f"  Scenarios: {self._n_scenarios} \n"
                  f"  Features: {self.n_features} \n"
                  f"    - Random: {self.n_random_features()} (with std > 0) \n"
                  f"    - Numeric: {self._n_num_features} \n"
                  f"    - Categorical: {self._n_cat_features} \n"
                  f"  Clustering methods: {self._methods_available}\n\n")
        
        string += "Clustering performed:\n"
        for method in self._methods_available:
            string += (f"  {method}: {self.n_runs[method]}\n")
            
        string += ("\nReference solution: \n"
                  f"  {self._reference_solution} \n")
        return string
                  
    def print_results(self, method=None, cardinality=None, index_sample=None, return_string=False):
        # limit print of the weights
        np.set_printoptions(threshold=15, linewidth=120) # max 15 weights
        # get methods, cardinality and indices to print
        methods = self._methods_available if method is None else [method]
        is_card_valid = lambda card: True if cardinality is None else card == cardinality
        is_index_valid = lambda index: True if index_sample is None else index == index_sample
        string = f"Reference solution: \n  {self._reference_solution} \n"
        for method in methods:
            string += f"\n{method}: \n" + "#" * (len(method)+1) + "\n"
            for card in self.cardinalities(method):
                if not is_card_valid(card): continue
                string += f"  Cardinality: {card} \n  " + "-" * len(f"Cardinality: {card}") + "\n"
                for index in range(self.get_n_runs(method, card)):
                    if not is_index_valid(index): continue
                    key = (method, card, index)
                    string += (f"    Weights: {self.get_weights(*key)} \n"
                               f"    Representatives: {self.get_representatives(*key)} \n"
                               f"    Score: {self.get_score(*key)} \n"
                               f"    Param: {self.get_param(*key)} \n"
                                "    Random features: "
                                        f"{self.n_random_features(*key)}/{self.n_random_features()} \n"
                               f"    Time: {self.get_clust_time(*key):.2f} sec\n")                    
                    if self._solutions.get(key) is None:
                        string += (f"    Clustered problem: None \n"
                                   f"    Implement. problem: None \n"
                                   f"    Gap to ref. sol. (%): None\n\n")
                    else:
                        timelimits = self._solutions[key].keys()
                        timelimits_clust = [time for time in timelimits if isinstance(time, (int, float)) or time is None]
                        timelimits_tuple = [time for time in timelimits if isinstance(time, tuple)]
                        string += f"    Clustered problem: \n"
                        for timelimit in timelimits_clust:
                            string += f"      {timelimit} sec: {self.get_opt_solution(*key, timelimit)} \n"
                        string += f"    Implement. problem: \n"
                        for timelimit in timelimits_tuple:
                            string += f"      {timelimit} sec: {self.get_eval_solution(*key, timelimit)} \n"
                        string += f"    Gap to ref. sol. (%): \n"
                        for timelimit in timelimits_tuple:
                            string += f"      {timelimit} sec: {self.get_gap(*key, timelimit)} \n"
                    string += "\n"
        if return_string:
            return string
        else:
            print(string, end="")
            
    # --- Sanity check ---
    def _check_opportunity_cost_matrix(self):
        if self._opport_cost_matrix is None:
            return 
        assert self._opport_cost_matrix.shape == (self._n_scenarios, self._n_scenarios), \
            ("Shape of opportunity cost matrix doesn't match the number of scenarios: "
             f"{self._opport_cost_matrix.shape} != {(self._n_scenarios, self._n_scenarios)}")
            
    def _check_sanity(self):
        self._check_opportunity_cost_matrix()
        assert len(self._scenarios) >= 3, \
            f"Clustering the scenarios makes sense for 3 scenarios or more, not {len(self._scenarios)}."